"""
I once used the help of this module to implement a basic compositing manager
using xpyb. It only did transparency, and not well. (Performance was actually
quite nice.) It's called pycompmgr.

This is meant to be a very close translation to the corresponding xcb-util
module. Mostly because I lack a deep understanding of everything here.
"""
from xpybutil.compat import render

class PictFormat:
    Id = 1
    Type = 2
    Depth = 4
    Red = 8
    RedMask = 16
    Green = 32
    GreenMask = 64
    Blue = 128
    BlueMask = 256
    Alpha = 512
    AlphaMask = 1024
    Colormap = 2048

class PictStandard:
    Argb32 = 0
    Rgb24 = 1
    A8 = 2
    A4 = 3
    A1 = 4

standardFormats = [
    # ARGB_32
    {
        'template': {
            'id': 0,
            'type': render.PictType.Direct,
            'depth': 32,
            'direct': {
                'red': 16,
                'red_mask': 0xff,
                'green': 8,
                'green_mask': 0xff,
                'blue': 0,
                'blue_mask': 0xff,
                'alpha': 24,
                'alpha_mask': 0xff
            },
            'colormap': 0
        },

        'mask': (PictFormat.Type | PictFormat.Depth | PictFormat.Red |
                 PictFormat.RedMask | PictFormat.Green | PictFormat.GreenMask |
                 PictFormat.Blue | PictFormat.BlueMask | PictFormat.Alpha |
                 PictFormat.AlphaMask)
    },

    # RGB_24
    {
        'template': {
            'id': 0,
            'type': render.PictType.Direct,
            'depth': 24,
            'direct': {
                'red': 16,
                'red_mask': 0xff,
                'green': 8,
                'green_mask': 0xff,
                'blue': 0,
                'blue_mask': 0xff,
                'alpha': 0,
                'alpha_mask': 0x00
            },
            'colormap': 0
        },

        'mask': (PictFormat.Type | PictFormat.Depth | PictFormat.Red |
                 PictFormat.RedMask | PictFormat.Green | PictFormat.GreenMask |
                 PictFormat.Blue | PictFormat.BlueMask | PictFormat.AlphaMask)
    },

    # A_8
    {
        'template': {
            'id': 0,
            'type': render.PictType.Direct,
            'depth': 8,
            'direct': {
                'red': 0,
                'red_mask': 0x00,
                'green': 0,
                'green_mask': 0x00,
                'blue': 0,
                'blue_mask': 0x00,
                'alpha': 0,
                'alpha_mask': 0xff
            },
            'colormap': 0
        },

        'mask': (PictFormat.Type | PictFormat.Depth | PictFormat.RedMask |
                 PictFormat.GreenMask | PictFormat.BlueMask |
                 PictFormat.Alpha | PictFormat.AlphaMask)
    },

    # A_4
    {
        'template': {
            'id': 0,
            'type': render.PictType.Direct,
            'depth': 4,
            'direct': {
                'red': 0,
                'red_mask': 0x00,
                'green': 0,
                'green_mask': 0x00,
                'blue': 0,
                'blue_mask': 0x00,
                'alpha': 0,
                'alpha_mask': 0x0f
            },
            'colormap': 0
        },

        'mask': (PictFormat.Type | PictFormat.Depth | PictFormat.RedMask |
                 PictFormat.GreenMask | PictFormat.BlueMask |
                 PictFormat.Alpha | PictFormat.AlphaMask)
    },

    # A_1
    {
        'template': {
            'id': 0,
            'type': render.PictType.Direct,
            'depth': 1,
            'direct': {
                'red': 0,
                'red_mask': 0x00,
                'green': 0,
                'green_mask': 0x00,
                'blue': 0,
                'blue_mask': 0x00,
                'alpha': 0,
                'alpha_mask': 0x01
            },
            'colormap': 0
        },

        'mask': (PictFormat.Type | PictFormat.Depth | PictFormat.RedMask |
                 PictFormat.GreenMask | PictFormat.BlueMask |
                 PictFormat.Alpha | PictFormat.AlphaMask)
    }
]

def find_visual_format(formats, visid):
    for screen in formats.screens:
        for depth in screen.depths:
            for visual in depth.visuals:
                if visual.visual == visid:
                    return visual

def find_format(formats, mask, template, count):
    if not formats:
        return None

    for format in formats.formats:
        if mask & PictFormat.Id:
            if template['id'] != format.id:
                continue
        if mask & PictFormat.Type:
            if template['type'] != format.type:
                continue
        if mask & PictFormat.Depth:
            if template['depth'] != format.depth:
                continue
        if mask & PictFormat.Red:
            if template['direct']['red'] != format.direct.red_shift:
                continue
        if mask & PictFormat.RedMask:
            if template['direct']['red_mask'] != format.direct.red_mask:
                continue
        if mask & PictFormat.Green:
            if template['direct']['green'] != format.direct.green_shift:
                continue
        if mask & PictFormat.GreenMask:
            if template['direct']['green_mask'] != format.direct.green_mask:
                continue
        if mask & PictFormat.Blue:
            if template['direct']['blue'] != format.direct.blue_shift:
                continue
        if mask & PictFormat.BlueMask:
            if template['direct']['blue_mask'] != format.direct.blue_mask:
                continue
        if mask & PictFormat.Alpha:
            if template['direct']['alpha'] != format.direct.alpha_shift:
                continue
        if mask & PictFormat.AlphaMask:
            if template['direct']['alpha_mask'] != format.direct.alpha_mask:
                continue
        if mask & PictFormat.Colormap:
            if template['colormap'] != format.colormap:
                continue

        if not count:
            return format
        count -= 1

    return None

def find_standard_format(formats, format):
    if format < 0 or format >= len(standardFormats):
        return None

    return find_format(
        formats,
        standardFormats[format]['mask'],
        standardFormats[format]['template'],
        0
    )




"""
A vast assortment of utility functions. The ones of interest to
you are probably 'get_atom' and 'get_atom_name'. The rest are
heavily used throughout the rest of xpybutil.
"""
import struct
import sys

from xpybutil.compat import xproto

from xpybutil import conn

__atom_cache = {}
__atom_nm_cache = {}

class Cookie(object):
    """
    The base Cookie class. The role of a cookie is to serve as an intermediary
    between you and the X server. After calling one of the many functions
    in the ewmh or icccm modules, the X server is typically not contacted
    immediately. Usually, you'll need to call the 'check()' or 'reply()'
    methods on the cookie returned by one of the functions in the ewmh or
    icccm modules. (Alternatively, you could flush the X buffer using
    ``conn_obj.flush()``.)
    """
    def __init__(self, cookie):
        self.cookie = cookie

    def check(self):
        return self.cookie.check()

class PropertyCookie(Cookie):
    """
    A regular property cookie that uses 'get_property_value' to return a nicer
    version to you. (Instead of raw X data.)
    """
    def reply(self):
        return get_property_value(self.cookie.reply())

class PropertyCookieSingle(Cookie):
    """
    A cookie that should only be used when one is guaranteed a single logical
    result. Namely, 'get_property_value' will be stupid and return a single
    list. This class checks for that, and takes the head of that list.
    """
    def reply(self):
        ret = get_property_value(self.cookie.reply())

        if ret and isinstance(ret, list) and len(ret) == 1:
            return ret[0]
        return ret

class AtomCookie(Cookie):
    """
    Pulls the ATOM identifier out of the reply object.
    """
    def reply(self):
        return self.cookie.reply().atom

class AtomNameCookie(Cookie):
    """
    Converts the null terminated list of characters (that represents an
    ATOM name) to a string.
    """
    def reply(self):
        return bytes(self.cookie.reply().name.buf()).decode('utf-8')

def get_property_value(property_reply):
    """
    A function that takes a property reply object, and turns its value into
    something nice for us.

    In particular, if the format of the reply is '8', then assume that it
    is a string. Moreover, it could be a list of strings that are null
    terminated. In this case, return a list of Python strings. Otherwise, just
    convert it to a string and remove the null terminator if it exists.

    Otherwise, the format must be a list of integers that has to be unpacked.
    Sometimes, these integers are ATOM identifiers, so it is useful to map
    'get_atom_name' over this list if that's the case.

    :param property_reply: An object returned by a cookie's "reply" method.
    :type property_reply: xcb.xproto.GetPropertyReply
    :return: Either a string, a list of strings or a list of integers depending
             upon the format of the property reply.
    """
    if property_reply.format == 8:
        ret = bytes(property_reply.value.buf()).split(b'\0')
        if ret[-1] == '': ret.pop()
        ret = [ x.decode('utf-8') for x in ret ]
        return ret[0] if len(ret) == 1 else ret
    elif property_reply.format in (16, 32):
        return list(struct.unpack('I' * property_reply.value_len,
                                  property_reply.value.buf()))

    return None

def get_property(window, atom):
    """
    Abstracts the process of issuing a GetProperty request.

    You'll typically want to call the ``reply`` method on the return value of
    this function, and pass that result to
    'get_property_value' so that the data is nicely formatted.

    :param window: A window identifier.
    :type window: int
    :param atom: An atom identifier.
    :type atom: int OR str
    :rtype: xcb.xproto.GetPropertyCookie
    """
    stringtype = str if sys.version_info[0] >= 3 else basestring
    if isinstance(atom, stringtype):
        atom = get_atom(atom)
    return conn.core.GetProperty(False, window, atom,
                                 xproto.GetPropertyType.Any, 0,
                                 2 ** 32 - 1)

def get_property_unchecked(window, atom):
    """
    Abstracts the process of issuing a GetPropertyUnchecked request.

    You'll typically want to call the ``reply`` method on the return value of
    this function, and pass that result to
    'get_property_value' so that the data is nicely formatted.

    :param window: A window identifier.
    :type window: int
    :param atom: An atom identifier.
    :type atom: int OR str
    :rtype: xcb.xproto.GetPropertyCookie
    """
    stringtype = str if sys.version_info[0] >= 3 else basestring
    if isinstance(atom, stringtype):
        atom = get_atom(atom)
    return conn.core.GetPropertyUnchecked(False, window, atom,
                                          xproto.GetPropertyType.Any, 0,
                                          2 ** 32 - 1)

def build_atom_cache(atoms):
    """
    Quickly builds a cache of ATOM names to ATOM identifiers (and the reverse).
    You'll only need to use this function if you're using atoms not defined in
    the ewmh, icccm or motif modules. (Otherwise, those modules will build this
    cache for you.)

    The 'get_atom' and 'get_atom_name' function automatically use this cache.

    :param atoms: A list of atom names.
    :rtype: void
    """
    global __atom_cache, __atom_nm_cache

    if conn is None:
        return

    for atom in atoms:
        __atom_cache[atom] = __get_atom_cookie(atom, only_if_exists=False)
    for atom in __atom_cache:
        if isinstance(__atom_cache[atom], AtomCookie):
            __atom_cache[atom] = __atom_cache[atom].reply()

    __atom_nm_cache = dict((v, k) for k, v in __atom_cache.items())

def get_atom(atom_name, only_if_exists=False):
    """
    Queries the X server for an ATOM identifier using a name. If we've already
    cached the identifier, then we don't contact the X server.

    If the identifier is not cached, it is added to the cache.

    If 'only_if_exists' is false, then the atom is created if it does not exist
    already.

    :param atom_name: An atom name.
    :type atom_name: str
    :param only_if_exists: If false, the atom is created if it didn't exist.
    :type only_if_exists: bool
    :return: ATOM identifier.
    :rtype: int
    """
    global __atom_cache

    a = __atom_cache.setdefault(atom_name,
                                __get_atom_cookie(atom_name,
                                                  only_if_exists).reply())
    if isinstance(a, AtomCookie):
        a = a.reply()

    return a

def get_atom_name(atom):
    """
    Queries the X server for an ATOM name using the specified identifier.
    If we've already cached the name, then we don't contact the X server.

    If the atom name is not cached, it is added to the cache.

    :param atom: An atom identifier.
    :type atom: int
    :return: ATOM name.
    :rtype: str
    """
    global __atom_nm_cache

    a = __atom_nm_cache.setdefault(atom, __get_atom_name_cookie(atom).reply())

    if isinstance(a, AtomNameCookie):
        a = a.reply()

    return a

def __get_atom_cookie(atom_name, only_if_exists=False):
    """
    Private function that issues the xpyb call to intern an atom.

    :type atom_name: str
    :type only_if_exists: bool
    :rtype: xcb.xproto.InternAtomCookie
    """
    atom_bytes = atom_name.encode('ascii')
    atom = conn.core.InternAtomUnchecked(only_if_exists, len(atom_bytes),
                                         atom_bytes)
    return AtomCookie(atom)

def __get_atom_name_cookie(atom):
    """
    Private function that issues the xpyb call to get an ATOM identifier's name.

    :type atom: int
    :rtype: xcb.xproto.GetAtomNameCookie
    """
    return AtomNameCookie(conn.core.GetAtomNameUnchecked(atom))




try:
    import xcffib as xcb
    import xcffib.xproto as xproto
    import xcffib.xinerama as xinerama
    import xcffib.randr as randr
    import xcffib.render as render
    from xcffib import XcffibException as xcb_Exception
    from xcffib import ConnectionException as xcb_ConnectException

except ImportError:
    import xcb
    import xcb.xproto as xproto
    import xcb.xinerama as xinerama
    import xcb.randr as randr
    import xcb.render as render
    from xcb import Exception as xcb_Exception
    from xcb import ConnectException as xcb_ConnectException



"""
A module that can send client events to windows. It also allows
registering callback functions to particular events. It can also
run the main event loop.
"""
from collections import defaultdict, deque
import struct
import sys
import traceback

from xpybutil.compat import xcb_Exception, xproto

from xpybutil import conn, root, util

__queue = deque()
__callbacks = defaultdict(list)
EM = xproto.EventMask

stringtype = str if sys.version_info[0] >= 3 else basestring

class Event(object):
    KeyPressEvent = 2
    KeyReleaseEvent = 3
    ButtonPressEvent = 4
    ButtonReleaseEvent = 5
    MotionNotifyEvent = 6
    EnterNotifyEvent = 7
    LeaveNotifyEvent = 8
    FocusInEvent = 9
    FocusOutEvent = 10
    KeymapNotifyEvent = 11
    ExposeEvent = 12
    GraphicsExposureEvent = 13
    NoExposureEvent = 14
    VisibilityNotifyEvent = 15
    CreateNotifyEvent = 16
    DestroyNotifyEvent = 17
    UnmapNotifyEvent = 18
    MapNotifyEvent = 19
    MapRequestEvent = 20
    ReparentNotifyEvent = 21
    ConfigureNotifyEvent = 22
    ConfigureRequestEvent = 23
    GravityNotifyEvent = 24
    ResizeRequestEvent = 25
    CirculateNotifyEvent = 26
    CirculateRequestEvent = 27
    PropertyNotifyEvent = 28
    SelectionClearEvent = 29
    SelectionRequestEvent = 30
    SelectionNotifyEvent = 31
    ColormapNotifyEvent = 32
    ClientMessageEvent = 33
    MappingNotifyEvent = 34

def replay_pointer():
    conn.core.AllowEventsChecked(xproto.Allow.ReplayPointer,
                                 xproto.Time.CurrentTime).check()

def send_event(destination, event_mask, event, propagate=False):
    return conn.core.SendEvent(propagate, destination, event_mask, event)

def send_event_checked(destination, event_mask, event, propagate=False):
    return conn.core.SendEventChecked(propagate, destination, event_mask, event)

def pack_client_message(window, message_type, *data):
    assert len(data) <= 5

    if isinstance(message_type, stringtype):
        message_type = util.get_atom(message_type)

    data = list(data)
    data += [0] * (5 - len(data))

    # Taken from
    # http://xcb.freedesktop.org/manual/structxcb__client__message__event__t.html
    return struct.pack('BBH7I', Event.ClientMessageEvent, 32, 0, window,
                       message_type, *data)

def root_send_client_event(window, message_type, *data):
    mask = EM.SubstructureNotify | EM.SubstructureRedirect
    packed = pack_client_message(window, message_type, *data)
    return send_event(root, mask, packed)

def root_send_client_event_checked(window, message_type, *data):
    mask = EM.SubstructureNotify | EM.SubstructureRedirect
    packed = pack_client_message(window, message_type, *data)
    return send_event_checked(root, mask, packed)

def is_connected(event_name, window, callback):
    member = '%sEvent' % event_name
    assert hasattr(xproto, member)

    key = (getattr(xproto, member), window)
    return key in __callbacks and callback in __callbacks[key]

def connect(event_name, window, callback):
    member = '%sEvent' % event_name
    assert hasattr(xproto, member)

    key = (getattr(xproto, member), window)
    __callbacks[key].append(callback)

def disconnect(event_name, window):
    member = '%sEvent' % event_name
    assert hasattr(xproto, member)

    key = (getattr(xproto, member), window)
    if key in __callbacks:
        del __callbacks[key]

def read(block=False):
    if block:
        e = conn.wait_for_event()
        __queue.appendleft(e)

    while True:
        e = conn.poll_for_event()

        if not e:
            break

        __queue.appendleft(e)

def main():
    try:
        while True:
            read(block=True)
            for e in queue():
                w = None
                if isinstance(e, xproto.MappingNotifyEvent):
                    w = None
                elif isinstance(e, xproto.MapRequestEvent):
                    # Force all MapRequestEvents to go to the root window so
                    # a window manager using xpybutil can get them.
                    w = root
                elif hasattr(e, 'window'):
                    w = e.window
                elif hasattr(e, 'event'):
                    w = e.event
                elif hasattr(e, 'owner'):
                    w = e.owner
                elif hasattr(e, 'requestor'):
                    w = e.requestor

                key = (e.__class__, w)
                for cb in __callbacks.get(key, []):
                    cb(e)
    except xcb_Exception:
        traceback.print_exc()
        sys.exit(1)

def queue():
    while len(__queue):
        yield __queue.pop()

def peek():
    return list(__queue)




"""
Provides a listing of all built-in X cursors, and provides a
function 'create_font_cursor' to create one. This object can
then be directly used in a call to CreateWindow or
ChangeWindowAttributes.
"""
from xpybutil import conn

class FontCursor:
    """Constants for all X cursors. To be used in ``create_font_cursor``."""
    XCursor = 0
    Arrow = 2
    BasedArrowDown = 4
    BasedArrowUp = 6
    Boat = 8
    Bogosity = 10
    BottomLeftCorner = 12
    BottomRightCorner = 14
    BottomSide = 16
    BottomTee = 18
    BoxSpiral = 20
    CenterPtr = 22
    Circle = 24
    Clock = 26
    CoffeeMug = 28
    Cross = 30
    CrossReverse = 32
    Crosshair = 34
    DiamondCross = 36
    Dot = 38
    DotBoxMask = 40
    DoubleArrow = 42
    DraftLarge = 44
    DraftSmall = 46
    DrapedBox = 48
    Exchange = 50
    Fleur = 52
    Gobbler = 54
    Gumby = 56
    Hand1 = 58
    Hand2 = 60
    Heart = 62
    Icon = 64
    IronCross = 66
    LeftPtr = 68
    LeftSide = 70
    LeftTee = 72
    LeftButton = 74
    LLAngle = 76
    LRAngle = 78
    Man = 80
    MiddleButton = 82
    Mouse = 84
    Pencil = 86
    Pirate = 88
    Plus = 90
    QuestionArrow = 92
    RightPtr = 94
    RightSide = 96
    RightTee = 98
    RightButton = 100
    RtlLogo = 102
    Sailboat = 104
    SBDownArrow = 106
    SBHDoubleArrow = 108
    SBLeftArrow = 110
    SBRightArrow = 112
    SBUpArrow = 114
    SBVDoubleArrow = 116
    Shuttle = 118
    Sizing = 120
    Spider = 122
    Spraycan = 124
    Star = 126
    Target = 128
    TCross = 130
    TopLeftArrow = 132
    TopLeftCorner = 134
    TopRightCorner = 136
    TopSide = 138
    TopTee = 140
    Trek = 142
    ULAngle = 144
    Umbrella = 146
    URAngle = 148
    Watch = 150
    XTerm = 152

def create_font_cursor(cursor_id, fore_red=0, fore_green=0, fore_blue=0,
                       back_red=0xffff, back_green=0xffff, back_blue=0xffff):
    """
    Function to create cursor resources on the X server. You can use the
    return value of this function with mousebind.grab_pointer to make the
    mouse change when it's being grabbed.

    The rest of the parameters are a way to colorize the cursor.

    :param cursor_id: A numeric identifier for a cursor.
    :type cursor_id: A class variable of FontCursor
    :type fore_red: int
    :type fore_green: int
    :type fore_blue: int
    :type back_red: int
    :type back_green: int
    :type back_blue: int
    :return: A cursor X identifier.
    :rtype: int
    """
    font = conn.generate_id()
    cursor = conn.generate_id()

    conn.core.OpenFontChecked(font, len('cursor'), 'cursor').check()
    conn.core.CreateGlyphCursorChecked(cursor, font, font, cursor_id,
                                       cursor_id + 1, fore_red, fore_green,
                                       fore_blue, back_red, back_green,
                                       back_blue).check()
    conn.core.CloseFont(font)

    return cursor




"""
A few utility functions related to client windows. In
particular, getting an accurate geometry of a client window
including the decorations (this can vary with the window
manager). Also, a functon to move and/or resize a window
accurately by the top-left corner. (Also can change based on
the currently running window manager.)

This module also contains a function 'listen' that must be used
in order to receive certain events from a window. For example,
if you wanted to run 'func' whenever a property on the root
window changed, you could do something like:

  ::

    import xpybutil
    import xpybutil.event as event
    import xpybutil.ewmh as ewmh
    import xpybutil.util as util
    import xpybutil.window as window

    def func(e):
      if util.get_atom_name(e.atom) == '_NET_ACTIVE_WINDOW':
        # Do something whenever the active window changes
        active_window_id = ewmh.get_active_window().reply()

    window.listen(xpybutil.root, 'PropertyChange')
    event.connect('PropertyNotify', xpybutil.root, func)

The idea here is to tell X that you want events that fall under
the 'PropertyChange' category. Then you bind 'func' to the
particular event 'PropertyNotify'.
"""
from xpybutil.compat import xproto

from xpybutil import conn
import xpybutil.ewmh as ewmh

class WindowManagers(object):
    """
    A list of window managers that xpybutil is aware of.

    These mostly modify how we determine the size of a client. In particular,
    KWin's decorations are in the parent of the parent of the client, whereas
    in Openbox, they are simply in the parent of the client.

    I am not sure whether I have plans to expand this list.
    """
    Unknown = 0
    Openbox = 1
    KWin = 2

def listen(window, *event_mask_names):
    """
    Makes X report events for the masks provided.

    This function must be called in order to get X send you events about a
    particular window. (i.e., it is not simply enough to call 'event.connect'.)

    This page is required reading if you are to do any event processing:
    http://tronche.com/gui/x/xlib/events/processing-overview.html

    :param window: Window identifier.
    :type window: int
    :param event_mask_names: List of mask names.
    :type event_mask_names: List of xcb.xproto.EventMask class variable names
    :rtype: void
    """
    masks = 0
    for mask_name in event_mask_names:
        assert hasattr(xproto.EventMask, mask_name)
        masks |= getattr(xproto.EventMask, mask_name)

    conn.core.ChangeWindowAttributesChecked(window, xproto.CW.EventMask,
                                            [masks]).check()

def get_parent_window(window):
    """
    Uses QueryTree() to find the parent of the given window.

    :param window: Window identifier.
    :return: Parent window identifier of 'window'.
    :rtype: int
    """
    return conn.core.QueryTree(window).reply().parent

def get_geometry(window, window_manager=None):
    """
    Returns the geometry of a window. This function will do its best to get
    the real geometry of a window; typically by inspecting the window's
    decorations if there are any.

    Since decorations are different for each window manager, you'll get the
    best possible results by passing a supported window manager in.

    :param window: Window identifier.
    :param window_manager: A class variable from Window.WindowManagers
    :type window_manager: int
    :return: Real geometry of a client window starting from the top-left corner.
    :rtype: (top_left_x, top_left_y, width, height)
    """

    if window_manager is WindowManagers.KWin:
        p = get_parent_window(window)
        return __get_geometry(get_parent_window(p))
    else:
        return __get_geometry(get_parent_window(window))

def moveresize(win, x=None, y=None, w=None, h=None, window_manager=None):
    """
    This function attempts to properly move/resize a window, accounting for
    its decorations.

    It doesn't rely upon _NET_FRAME_EXTENTS, but instead, uses the actual
    parent window to adjust the width and height. (I've found _NET_FRAME_EXTENTS
    to be wildly unreliable.)

    :param win: Window identifier.
    :param x: Top left x coordinate.
    :param y: Top left y coordinate.
    :param w: Client width.
    :param h: Client height.
    :param window_manager: A class variable from Window.WindowManagers
    :type window_manager: int
    :rtype: void
    """
    if window_manager is WindowManagers.KWin:
        tomove = get_parent_window(get_parent_window(win))
    else:
        tomove = get_parent_window(win)

    if tomove:
        cx, cy, cw, ch = __get_geometry(win)
        px, py, pw, ph = __get_geometry(tomove)

        w -= pw - cw
        h -= ph - ch

    ewmh.request_moveresize_window(win, x=x, y=y, width=max(1, w),
                                   height=max(1, h), source=2)

def __get_geometry(win):
    """
    Private function that abstracts the low level GetGeometry call.

    If you're looking for the size of a window including its decorations,
    please use ``window.get_geometry``.

    :param win: Window identifier.
    :return: X rectangle of the window.
    :rtype: (x, y, width, height)
    """
    raw = conn.core.GetGeometry(win).reply()
    return raw.x, raw.y, raw.width, raw.height




"""
An incomplete and haphazard collection of functions that can
bridge a gap between PIL and drawing images with X.
"""
from xpybutil.compat import xproto

from xpybutil import conn
from PIL import Image

def net_wm_icon_to_bgra(data):
    ret = [0] * (len(data) * 4)
    for i, d in enumerate(data):
        alpha = d >> 24
        red = ((alpha << 24) ^ d) >> 16
        green = (((alpha << 24) + (red << 16)) ^ d) >> 8
        blue = (((alpha << 24) + (red << 16) + (green << 8)) ^ d) >> 0

        pixel = i * 4
        ret[pixel] = blue
        ret[pixel + 1] = green
        ret[pixel + 2] = red
        ret[pixel + 3] = alpha

    return ret

def color_humanize(clr):
    t = hex(clr).replace('0x', '')
    return '#%s' % ('0' * (6 - len(t)) + t)

def hex_to_rgb(h):
    s = color_humanize(h)
    return (int(s[1:3], 16), int(s[3:5], 16), int(s[5:7], 16))

def get_image_from_pixmap(pid):
    try:
        geom = conn.core.GetGeometry(pid).reply()
        pimg = conn.core.GetImage(xproto.ImageFormat.ZPixmap, pid,
                                  0, 0, geom.width, geom.height,
                                  2**32 - 1).reply().data

        return geom.width, geom.height, pimg
    except xproto.BadDrawable:
        return 0, 0, []

def get_image(width, height, data):
    im = Image.fromstring('RGBA', (width, height),
                          ''.join([chr(i) for i in data]), 'raw', 'BGRA')

    return im

def get_bitmap(width, height, data):
    im = Image.fromstring('1', (width, height),
                          ''.join([chr(i) for i in data]), 'raw', '1;R')

    return im

def get_data(image):
    return [ord(s) for s in image.tostring('raw', 'BGRA')]




"""
Implements the entire EWMH spec. The spec can be found here:
http://standards.freedesktop.org/wm-spec/wm-spec-latest.html

xpybutil's primary purpose was to make accessing ICCCM and EWMH related
information extremely easy. This can be done by using the ewmh or icccm
modules. Here's a quick example (assuming xpybutil is installed):

 ::

  import xpybutil.ewmh as ewmh

  desktop_names = ewmh.get_desktop_names().reply()
  current_desktop = ewmh.get_current_desktop().reply()

  print names.get(current_desktop, current_desktop)

This imports the ewmh module (and by extension, connects to the X server) and
fetches a list of the current desktop names and the current desktop index
(starting from 0). Since not every desktop must be named, the desktop index is
printed if it has no name.

Note that the functions in the ewmh and icccm module return *cookies*. In order
to pull a response from the X server, call the 'reply()' method on a cookie
object. To force a response sent to the X server, call the 'check()' method on
the corresponding cookie.

Much of the EWMH and ICCCM modules encapsulate packing data from convenient
Python data types into C structs (using the 'struct' Python module). For
example, if one wants to fetch the partial struts set by some window with
identifier ID, you could do:

 ::

  import xpybutil.ewmh as ewmh

  print ewmh.get_wm_strut_partial(ID).reply()

Which outputs a dictionary with 12 entries, where each corresponds to a value
in the partial strut. (i.e., 'left', 'top_end_x', 'right_start_y', etc...).

In general, particularly with the EWMH module, the ewmh module is very nearly
representative of the spec itself. Functions that get property values start
with ``get_``, functions that set property values start with ``set_``, and
functions that send an event to a client (which typically requests the window
manager to DO something) start with ``request_``.

If a request has no reply (typically the ``set_`` functions), then the
default is to call it 'unchecked'. If you want to check the result (and force
retrieval), then use the '_checked' variant.

The reverse goes for the ``get_`` functions. By default, they are checked, but
you can use the '_unchecked' variant too.

Basically, you should probably almost always use the 'checked' variant of
everything. The cases where you don't are when you want to send a whole bunch
of requests to the X server at once. You could use the unchecked invariant, and
after you've initialized all the cookies, calling 'flush' will force
communication with the X server.

Finally, unless you're writing a window manager or creating a client window
from scratch, you'll almost always want to use the ``get_`` and ``request_``
functions, not ``set_``. For example, if you want to change the current
desktop...

  DON'T DO: ``ewmh.set_current_desktop_checked(2).check()``

  DO:       ``ewmh.request_current_desktop_checked(2).check()``

The former will probably not work, but the latter will. Just follow the EWMH
spec :-)
"""
import struct

from xpybutil.compat import xproto

from xpybutil import conn as c, root, event, util

__atoms = [
   # Non-standard
   '_NET_VISIBLE_DESKTOPS', '_NET_WM_WINDOW_OPACITY',

   # Root Window Properties (and Related Messages)
   '_NET_SUPPORTED', '_NET_CLIENT_LIST', '_NET_CLIENT_LIST_STACKING',
   '_NET_NUMBER_OF_DESKTOPS', '_NET_DESKTOP_GEOMETRY',
   '_NET_DESKTOP_VIEWPORT', '_NET_CURRENT_DESKTOP',
   '_NET_DESKTOP_NAMES', '_NET_ACTIVE_WINDOW', '_NET_WORKAREA',
   '_NET_SUPPORTING_WM_CHECK', '_NET_VIRTUAL_ROOTS',
   '_NET_DESKTOP_LAYOUT', '_NET_SHOWING_DESKTOP',

   # Other Root Window Messages
   '_NET_CLOSE_WINDOW', '_NET_MORERESIZE_WINDOW', '_NET_WM_MOVERESIZE',
   '_NET_RESTACK_WINDOW', '_NET_REQUEST_FRAME_EXTENTS',

   # Application Window Properties
   '_NET_WM_NAME', '_NET_WM_VISIBLE_NAME', '_NET_WM_ICON_NAME',
   '_NET_WM_VISIBLE_ICON_NAME', '_NET_WM_DESKTOP',
   '_NET_WM_WINDOW_TYPE', '_NET_WM_STATE', '_NET_WM_ALLOWED_ACTIONS',
   '_NET_WM_STRUT', '_NET_WM_STRUT_PARTIAL', '_NET_WM_ICON_GEOMETRY',
   '_NET_WM_ICON', '_NET_WM_PID', '_NET_WM_HANDLED_ICONS',
   '_NET_WM_USER_TIME', '_NET_WM_USER_TIME_WINDOW',
   '_NET_FRAME_EXTENTS',

   # Window Manager Protocols
   '_NET_WM_PING', '_NET_WM_SYNC_REQUEST',
   '_NET_WM_FULLSCREEN_MONITORS',

   # Other Properties
   '_NET_WM_FULL_PLACEMENT',

   # _NET_WM_WINDOW_TYPE_*
   '_NET_WM_WINDOW_TYPE_DESKTOP', '_NET_WM_WINDOW_TYPE_DOCK',
   '_NET_WM_WINDOW_TYPE_TOOLBAR', '_NET_WM_WINDOW_TYPE_MENU',
   '_NET_WM_WINDOW_TYPE_UTILITY', '_NET_WM_WINDOW_TYPE_SPLASH',
   '_NET_WM_WINDOW_TYPE_DIALOG', '_NET_WM_WINDOW_TYPE_DROPDOWN_MENU',
   '_NET_WM_WINDOW_TYPE_POPUP_MENU', '_NET_WM_WINDOW_TYPE_TOOLTIP',
   '_NET_WM_WINDOW_TYPE_NOTIFICATION', '_NET_WM_WINDOW_TYPE_COMBO',
   '_NET_WM_WINDOW_TYPE_DND', '_NET_WM_WINDOW_TYPE_NORMAL',

   # _NET_WM_STATE_*
   '_NET_WM_STATE_MODAL', '_NET_WM_STATE_STICKY',
   '_NET_WM_STATE_MAXIMIZED_VERT', '_NET_WM_STATE_MAXIMIZED_HORZ',
   '_NET_WM_STATE_SHADED', '_NET_WM_STATE_SKIP_TASKBAR',
   '_NET_WM_STATE_SKIP_PAGER', '_NET_WM_STATE_HIDDEN',
   '_NET_WM_STATE_FULLSCREEN', '_NET_WM_STATE_ABOVE',
   '_NET_WM_STATE_BELOW', '_NET_WM_STATE_DEMANDS_ATTENTION',

   # _NET_WM_ACTION_*
   '_NET_WM_ACTION_MOVE', '_NET_WM_ACTION_RESIZE',
   '_NET_WM_ACTION_MINIMIZE', '_NET_WM_ACTION_SHADE',
   '_NET_WM_ACTION_STICK', '_NET_WM_ACTION_MAXIMIZE_HORZ',
   '_NET_WM_ACTION_MAXIMIZE_VERT', '_NET_WM_ACTION_FULLSCREEN',
   '_NET_WM_ACTION_CHANGE_DESKTOP', '_NET_WM_ACTION_CLOSE',
   '_NET_WM_ACTION_ABOVE', '_NET_WM_ACTION_BELOW',

   # Additional TEXT Type
   'UTF8_STRING'
]

class Orientation:
    Horz = 0
    Vert = 1

class StartingCorner:
    TopLeft = 0
    TopRight = 1
    BottomRight = 2
    BottomLeft = 3

class MoveResize:
    SizeTopLeft = 0
    SizeTop = 1
    SizeTopRight = 2
    SizeRight = 3
    SizeBottomRight = 4
    SizeBottom = 5
    SizeBottomLeft = 6
    SizeLeft = 7
    Move = 8
    SizeKeyboard = 9
    MoveKeyboard = 10
    Cancel = 11

class State:
    Remove = 0
    Add = 1
    Toggle = 2

# Some aliases
atom = util.get_atom
revent = event.root_send_client_event
revent_checked = event.root_send_client_event_checked
ATOM = xproto.Atom.ATOM
CARDINAL = xproto.Atom.CARDINAL
WINDOW = xproto.Atom.WINDOW

# Build the atom cache for quicker access
util.build_atom_cache(__atoms)


# _NET_SUPPORTED

def get_supported():
    """
    Returns a list of hints supported by the window manager.

    :return:        A list of atoms in the _NET_SUPPORTED property.
    :rtype:         util.PropertyCookie (ATOM[]/32)
    """
    return util.PropertyCookie(util.get_property(root, '_NET_SUPPORTED'))

def get_supported_unchecked():
    return util.PropertyCookie(util.get_property_unchecked(root,
                                                           '_NET_SUPPORTED'))

def set_supported(atoms):
    """
    Sets the list of hints supported by the window manager.

    :param atoms:   A list of atom identifiers.
    :type atoms:    ATOM[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(atoms), *atoms)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_SUPPORTED'), ATOM, 32, len(atoms),
                                 packed)

def set_supported_checked(atoms):
    packed = struct.pack('I' * len(atoms), *atoms)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_SUPPORTED'),
                                        ATOM, 32, len(atoms), packed)

# _NET_CLIENT_LIST

def get_client_list():
    """
    Returns a list of windows managed by the window manager.

    :return:        A list of window identifiers.
    :rtype:         util.PropertyCookie (ATOM[]/32)
    """
    return util.PropertyCookie(util.get_property(root, '_NET_CLIENT_LIST'))

def get_client_list_unchecked():
    return util.PropertyCookie(util.get_property_unchecked(root,
                                                           '_NET_CLIENT_LIST'))

def set_client_list(windows):
    """
    Sets the list of windows managed by the window manager.

    :param windows: A list of atom identifiers.
    :type windows:  ATOM[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(windows), *windows)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_CLIENT_LIST'),
                                 WINDOW, 32, len(windows), packed)

def set_client_list_checked(windows):
    packed = struct.pack('I' * len(windows), *windows)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_CLIENT_LIST'),
                                        WINDOW, 32, len(windows), packed)

# _NET_CLIENT_LIST_STACKING

def get_client_list_stacking():
    """
    Returns the window stacking order.

    :return:        A list of window identifiers.
    :rtype:         util.PropertyCookie (ATOM[]/32)
    """
    return util.PropertyCookie(util.get_property(root,
                                                 '_NET_CLIENT_LIST_STACKING'))

def get_client_list_stacking_unchecked():
    cook = util.get_property_unchecked(root, '_NET_CLIENT_LIST_STACKING')
    return util.PropertyCookie(cook)

def set_client_list_stacking(windows):
    """
    Sets the window stacking order.

    :param windows: A list of atom identifiers.
    :type windows:  ATOM[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(windows), *windows)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_CLIENT_LIST_STACKING'),
                                 WINDOW, 32, len(windows), packed)

def set_client_list_stacking_checked(windows):
    packed = struct.pack('I' * len(windows), *windows)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                 atom('_NET_CLIENT_LIST_STACKING'),
                                 WINDOW, 32, len(windows), packed)

# _NET_NUMBER_DESKTOPS

def get_number_of_desktops():
    """
    Returns the number of virtual desktops.

    :return:        The number of desktops.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    cook = util.get_property(root, '_NET_NUMBER_OF_DESKTOPS')
    return util.PropertyCookieSingle(cook)

def get_number_of_desktops_unchecked():
    cook = util.get_property_unchecked(root, '_NET_NUMBER_OF_DESKTOPS')
    return util.PropertyCookieSingle(cook)

def set_number_of_desktops(number_of_desktops):
    """
    Sets the number of desktops.

    :param number_of_desktops:  The number of desktops.
    :type number_of_desktops:   CARDINAL/32
    :rtype:                     xcb.VoidCookie
    """
    packed = struct.pack('I', number_of_desktops)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_NUMBER_OF_DESKTOPS'), CARDINAL, 32,
                                 1, packed)

def set_number_of_desktops_checked(number_of_desktops):
    packed = struct.pack('I', number_of_desktops)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_NUMBER_OF_DESKTOPS'),
                                        CARDINAL, 32, 1, packed)

def request_number_of_desktops(number_of_desktops):
    """
    Sends event to root window to set the number of desktops.

    :param number_of_desktops:  The number of desktops.
    :type number_of_desktops:   CARDINAL/32
    :rtype:                     xcb.VoidCookie
    """
    return revent(root, '_NET_NUMBER_OF_DESKTOPS', number_of_desktops)

def request_number_of_desktops_checked(number_of_desktops):
    return revent_checked(root, '_NET_NUMBER_OF_DESKTOPS', number_of_desktops)

# _NET_DESKTOP_GEOMETRY

class DesktopGeometryCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'width': v[0],
            'height': v[1]
        }

def get_desktop_geometry():
    """Returns the desktop geometry.

    :return:        A desktop geometry dictionary.
                    Keys: width, height
    :rtype:         DesktopGeometryCookie (CARDINAL[2]/32)
    """
    return DesktopGeometryCookie(util.get_property(root,
                                                   '_NET_DESKTOP_GEOMETRY'))

def get_desktop_geometry_unchecked():
    cook = util.get_property_unchecked(root, '_NET_DESKTOP_GEOMETRY')
    return DesktopGeometryCookie(cook)

def set_desktop_geometry(width, height):
    """
    Sets the desktop geometry.

    :param width:               The width of the desktop.
    :type width:                CARDINAL/32
    :param height:              The height of the desktop.
    :type height:               CARDINAL/32
    :rtype:                     xcb.VoidCookie
    """
    packed = struct.pack('II', width, height)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_DESKTOP_GEOMETRY'), CARDINAL, 32, 2,
                                 packed)

def set_desktop_geometry_checked(width, height):
    packed = struct.pack('II', width, height)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_DESKTOP_GEOMETRY'),
                                        CARDINAL, 32, 2, packed)

def request_desktop_geometry(width, height):
    """
    Sends event to root window to set the desktop geometry.

    :param width:               The width of the desktop.
    :type width:                CARDINAL/32
    :param height:              The height of the desktop.
    :type height:               CARDINAL/32
    :rtype:                     xcb.VoidCookie
    """
    return revent(root, '_NET_DESKTOP_GEOMETRY', width, height)

def request_desktop_geometry_checked(width, height):
    return revent_checked(root, '_NET_DESKTOP_GEOMETRY', width, height)

# _NET_DESKTOP_VIEWPORT

class DesktopViewportCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        ret = []
        for i in xrange(0, len(v), 2):
            ret.append({
                'x': v[i],
                'y': v[i + 1]
            })

        return ret

def get_desktop_viewport():
    """
    Returns x,y pairs defining the top-left corner of each desktop's viewport.

    :return:        A list of desktop viewport dictionaries.

                    Keys: x, y
    :rtype:         DesktopViewportCookie (CARDINAL[][2]/32)
    """
    return DesktopViewportCookie(util.get_property(root,
                                                   '_NET_DESKTOP_VIEWPORT'))

def get_desktop_viewport_unchecked():
    cook = util.get_property_unchecked(root, '_NET_DESKTOP_VIEWPORT')
    return DesktopViewportCookie(cook)

def set_desktop_viewport(pairs):
    """
    Sets the x,y pairs defining the top-left corner of each desktop's viewport.

    Array of pairs of cardinals that define the top left corner of each
    desktop's viewport. For Window Managers that don't support large
    desktops, this MUST always be set to (0,0).

    The Window Manager MAY choose to ignore this message, in which case
    _NET_DESKTOP_VIEWPORT property will remain unchanged.

    Pairs should look like:

        >>> [
        {'x': 0, 'y': 0},
        {'x': 500, 'y': 500},
        ...
        ]

    Which would set desktop 0's viewport top-left corner to 0,0 and desktop
    1's viewport top-left corner to 500,500.

    :param pairs:               A list of x,y dictionary pairs.
    :type pairs:                CARDINAL[][2]/32
    :rtype:                     xcb.VoidCookie
    """
    flatten = []
    for pair in pairs:
        flatten.append(pair['x'])
        flatten.append(pair['y'])

    packed = struct.pack('I' * len(flatten), *flatten)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_DESKTOP_VIEWPORT'),
                                 CARDINAL, 32, len(flatten), packed)

def set_desktop_viewport_checked(pairs):
    flatten = []
    for pair in pairs:
        flatten.append(pair['x'])
        flatten.append(pair['y'])

    packed = struct.pack('I' * len(flatten), *flatten)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_DESKTOP_VIEWPORT'),
                                        CARDINAL, 32, len(flatten), packed)

def request_desktop_viewport(x, y):
    """
    Sends event to root window to set the viewport position of current desktop.

    :param x:       The x position of the top-left corner.
    :type x:        CARDINAL/32
    :param y:       The y position of the top-left corner.
    :type y:        CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    return revent(root, '_NET_DESKTOP_VIEWPORT', x, y)

def request_desktop_viewport_checked(x, y):
    return revent_checked(root, '_NET_DESKTOP_VIEWPORT', x, y)

# _NET_CURRENT_DESKTOP

def get_current_desktop():
    """
    Returns the current desktop number.

    :return:        The index of the current desktop.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    return util.PropertyCookieSingle(util.get_property(root,
                                                       '_NET_CURRENT_DESKTOP'))

def get_current_desktop_unchecked():
    cook = util.get_property_unchecked(root, '_NET_CURRENT_DESKTOP')
    return util.PropertyCookieSingle(cook)

def set_current_desktop(current_desktop):
    """
    Sets the current desktop number.

    :param current_desktop:     The current desktop index.
    :type current_desktop:      CARDINAL/32
    :rtype:                     xcb.VoidCookie
    """
    packed = struct.pack('I', current_desktop)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_CURRENT_DESKTOP'), CARDINAL, 32, 1,
                                 packed)

def set_current_desktop_checked(current_desktop):
    packed = struct.pack('I', current_desktop)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_CURRENT_DESKTOP'),
                                        CARDINAL, 32, 1, packed)

def request_current_desktop(desktop_number,
                            timestamp=xproto.Time.CurrentTime):
    """
    Sends event to root window to set the current desktop.

    :param desktop_number:      The current desktop index.
    :type desktop_number:       CARDINAL/32
    :type timestamp:            Milliseconds.
    :rtype:                     xcb.VoidCookie
    """
    return revent(root, '_NET_CURRENT_DESKTOP', desktop_number, timestamp)

def request_current_desktop_checked(desktop_number,
                                    timestamp=xproto.Time.CurrentTime):
    return revent_checked(root, '_NET_CURRENT_DESKTOP',
                          desktop_number, timestamp)

# _NET_VISIBLE_DESKTOPS

def get_visible_desktops():
    """
    Returns a list of visible desktops.

    The first desktop is on Xinerama screen 0, the second is on Xinerama
    screen 1, etc.

    :return:        A list of visible desktops.
    :rtype:         util.PropertyCookie (CARDINAL[]/32)
    """
    return util.PropertyCookie(util.get_property(root, '_NET_VISIBLE_DESKTOPS'))

def get_visible_desktops_unchecked():
    cook = util.get_property_unchecked(root, '_NET_VISIBLE_DESKTOPS')
    return util.PropertyCookie(cook)

def set_visible_desktops(desktops):
    """
    Sets the list of visible desktops.

    :param desktops: A list of desktops.
    :type desktops:  CARDINAL[]/32
    :rtype:          xcb.VoidCookie
    """
    packed = struct.pack('I' * len(desktops), *desktops)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_VISIBLE_DESKTOPS'),
                                 CARDINAL, 32, len(desktops), packed)

def set_visible_desktops_checked(desktops):
    packed = struct.pack('I' * len(desktops), *desktops)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_VISIBLE_DESKTOPS'),
                                        CARDINAL, 32, len(desktops), packed)

# _NET_DESKTOP_NAMES

def get_desktop_names():
    """
    Returns a list of names of the virtual desktops.

    :return:        A list of virutal desktop names.
    :rtype:         util.PropertyCookie (UTF8_STRING[])
    """
    return util.PropertyCookie(util.get_property(root, '_NET_DESKTOP_NAMES'))

def get_desktop_names_unchecked():
    cook = util.get_property_unchecked(root, '_NET_DESKTOP_NAMES')
    return util.PropertyCookie(cook)

def set_desktop_names(desktop_names):
    """
    Sets the current list of desktop names.

    :param desktop_names:   A list of new desktop names.
    :type desktop_names:    UTF8_STRING[]
    :rtype:                 xcb.VoidCookie
    """
    # Null terminate the list of desktop names
    nullterm = []
    for desktop_name in desktop_names:
        nullterm.append(desktop_name + chr(0))
    nullterm = ''.join(nullterm)

    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_DESKTOP_NAMES'),
                                 atom('UTF8_STRING'), 8,
                                 len(nullterm), nullterm)

def set_desktop_names_checked(desktop_names):
    # Null terminate the list of desktop names
    nullterm = []
    for desktop_name in desktop_names:
        nullterm.append(desktop_name + chr(0))
    nullterm = ''.join(nullterm)

    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_DESKTOP_NAMES'),
                                        atom('UTF8_STRING'), 8,
                                        len(nullterm), nullterm)

# _NET_ACTIVE_WINDOW

def get_active_window():
    """
    Returns the identifier of the currently active window.

    :return:        The window ID of the active window.
    :rtype:         util.PropertyCookieSingle (WINDOW/32)
    """
    return util.PropertyCookieSingle(util.get_property(root,
                                                       '_NET_ACTIVE_WINDOW'))

def get_active_window_unchecked():
    cook = util.get_property_unchecked(root, '_NET_ACTIVE_WINDOW')
    return util.PropertyCookieSingle(cook)

def set_active_window(active):
    """
    Sets the identifier of the currently active window.

    :param active:  The identifier of the window that is active.
    :type active:   WINDOW/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I', active)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_ACTIVE_WINDOW'),
                                 WINDOW, 32, 1, packed)

def set_active_window_checked(active):
    packed = struct.pack('I', active)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_ACTIVE_WINDOW'),
                                        WINDOW, 32, 1, packed)

def request_active_window(active, source=1,
                          timestamp=xproto.Time.CurrentTime,
                          current=0):
    """
    Sends event to root window to set the active window.

    :param active:      The window ID of the window to make active.
    :type active:       WINDOW/32
    :param source:      The source indication.
    :type timestamp:    Milliseconds
    :param current:     Client's active toplevel window
    :rtype:             xcb.VoidCookie
    """
    return revent(active, '_NET_ACTIVE_WINDOW', source, timestamp, current)

def request_active_window_checked(active, source=1,
                                  timestamp=xproto.Time.CurrentTime,
                                  current=0):
    return revent_checked(active, '_NET_ACTIVE_WINDOW', source, timestamp,
                          current)

# _NET_WORKAREA

class WorkareaCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        ret = []
        for i in xrange(0, len(v), 4):
            ret.append({
                'x': v[i],
                'y': v[i + 1],
                'width': v[i + 2],
                'height': v[i + 3]
            })

        return ret

def get_workarea():
    """
    Returns the x, y, width and height defining the desktop workarea.

    :return:        A list of workarea dictionaries.

                    Keys: x, y, width, height
    :rtype:         util.WorkareaCookie (CARDINAL[][4]/32)
    """
    return WorkareaCookie(util.get_property(root, '_NET_WORKAREA'))

def get_workarea_unchecked():
    return WorkareaCookie(util.get_property_unchecked(root, '_NET_WORKAREA'))

def set_workarea(workareas):
    """
    Sets the workarea (x, y, width, height) for each desktop.

    :param workareas:   A list of x,y,width,height dictionaries.
    :type workareas:    CARDINAL[][4]/32
    :rtype:             xcb.VoidCookie
    """
    flatten = []
    for workarea in workareas:
        flatten.append(workarea['x'])
        flatten.append(workarea['y'])
        flatten.append(workarea['width'])
        flatten.append(workarea['height'])
    packed = struct.pack('I' * len(flatten), *flatten)

    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_WORKAREA'), CARDINAL, 32,
                                 len(flatten), packed)

def set_workarea_checked(workareas):
    flatten = []
    for workarea in workareas:
        flatten.append(workarea['x'])
        flatten.append(workarea['y'])
        flatten.append(workarea['width'])
        flatten.append(workarea['height'])
    packed = struct.pack('I' * len(flatten), *flatten)

    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_WORKAREA'), CARDINAL, 32,
                                        len(flatten), packed)

# _NET_SUPPORTING_WM_CHECK

def get_supporting_wm_check(wid):
    """
    Returns the identifier of the child window created by the window manager.

    :param wid:     The identifier of the window with the property.
    :type wid:      WINDOW/32
    :return:        The window ID of the child window.
    :rtype:         util.PropertyCookieSingle (WINDOW/32)
    """
    cook = util.get_property(wid, '_NET_SUPPORTING_WM_CHECK')
    return util.PropertyCookieSingle(cook)

def get_supporting_wm_check_unchecked(wid):
    cook = util.get_property_unchecked(wid, '_NET_SUPPORTING_WM_CHECK')
    return util.PropertyCookieSingle(cook)

def set_supporting_wm_check(wid, child):
    """
    Sets the identifier of the child window created by the window manager.

    :param wid:     The identifier of the window with the property.
    :type wid:      WINDOW/32
    :param child:   The identifier of the child window.
    :type child:    WINDOW/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I', child)
    return c.core.ChangeProperty(xproto.PropMode.Replace, wid,
                                 atom('_NET_SUPPORTING_WM_CHECK'),
                                 WINDOW, 32, 1, packed)

def set_supporting_wm_check_checked(wid, child):
    packed = struct.pack('I', child)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, wid,
                                        atom('_NET_SUPPORTING_WM_CHECK'),
                                        WINDOW, 32, 1, packed)

# _NET_VIRTUAL_ROOTS

def get_virtual_roots():
    """
    Returns a list of identifiers for the virtual root windows.

    :return:        A list of window identifiers for the virtual root windows.
    :rtype:         util.PropertyCookie (WINDOW[]/32)
    """
    return util.PropertyCookie(util.get_property(root, '_NET_VIRTUAL_ROOTS'))

def get_virtual_roots_unchecked():
    cook = util.get_property_unchecked(root, '_NET_VIRTUAL_ROOTS')
    return util.PropertyCookie(cook)

def set_virtual_roots(vroots):
    """
    Sets the identifiers of the virtual root windows.

    :param vroots:  A list of window identifiers for the virtual root windows.
    :type vroots:   WINDOW[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(vroots), *vroots)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_VIRTUAL_ROOTS'),
                                 WINDOW, 32, len(vroots), packed)

def set_virtual_roots_checked(vroots):
    packed = struct.pack('I' * len(vroots), *vroots)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_VIRTUAL_ROOTS'),
                                        WINDOW, 32, len(vroots), packed)

# _NET_DESKTOP_LAYOUT

class DesktopLayoutCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'orientation': v[0],
            'columns': v[1],
            'rows': v[2],
            'starting_corner': v[3] if len(v) > 3 else StartingCorner.TopLeft
        }

def get_desktop_layout():
    """
    Returns the desktop layout.

    :return:        A desktop layout dictionary.

                    Keys: orientation, columns, rows, starting_corner
    :rtype:         DesktopLayoutCookie (CARDINAL[4]/32)
    """
    return DesktopLayoutCookie(util.get_property(root, '_NET_DESKTOP_LAYOUT'))

def get_desktop_layout_unchecked():
    cook = util.get_property_unchecked(root, '_NET_DESKTOP_LAYOUT')
    return DesktopLayoutCookie(cook)

def set_desktop_layout(orientation, columns, rows,
                       starting_corner=StartingCorner.TopLeft):
    """
    Sets the desktop layout.

    :param orientation:         Horizontal or vertical orientation.
    :type orientation:          CARDINAL/32
    :param columns:             Number of columns.
    :type columns:              CARDINAL/32
    :param rows:                Number of rows.
    :type rows:                 CARDINAL/32
    :param starting_corner:     Top left, top right, bottom right, or bottom
                                left may be specified as a starting corner.
    :type starting_corner:      CARDINAL/32
    :rtype:                     xcb.VoidCookie
    """
    packed = struct.pack('IIII', orientation, columns, rows, starting_corner)
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_DESKTOP_LAYOUT'),
                                 CARDINAL, 32, 4, packed)

def set_desktop_layout_checked(orientation, columns, rows,
                               starting_corner=StartingCorner.TopLeft):
    packed = struct.pack('IIII', orientation, columns, rows, starting_corner)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_DESKTOP_LAYOUT'),
                                        CARDINAL, 32, 4, packed)

# _NET_SHOWING_DESKTOP

class ShowingDesktopCookie(util.PropertyCookieSingle):
    def reply(self):
        v = util.PropertyCookieSingle.reply(self)

        if v is None:
            return None

        if v == 1:
            return True
        return False

def get_showing_desktop():
    """
    Returns whether the window manager is in "showing the desktop" mode.

    :return:        Boolean whether the window manager is in "showing desktop"
                    mode or not.
    :rtype:         ShowingDesktopCookie (CARDINAL/32)
    """
    return ShowingDesktopCookie(util.get_property(root, '_NET_SHOWING_DESKTOP'))

def get_showing_desktop_unchecked():
    cook = util.get_property_unchecked(root, '_NET_SHOWING_DESKTOP')
    return ShowingDesktopCookie(cook)

def set_showing_desktop(showing_desktop):
    """
    Sets whether the window is in "showing the desktop" mode.

    :param showing_desktop:  Boolean whether the window manager is in "showing
                             desktop" mode or not.
    :type showing_desktop:   CARDINAL/32
    :rtype:                  xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, root,
                                 atom('_NET_SHOWING_DESKTOP'), CARDINAL, 32, 1,
                                 [showing_desktop])

def set_showing_desktop_checked(showing_desktop):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, root,
                                        atom('_NET_SHOWING_DESKTOP'),
                                        CARDINAL, 32, 1, [showing_desktop])

def request_showing_desktop(showing_desktop):
    """
    Sends event to root window to put window manager in "showing desktop" mode.

    :param showing_desktop:  Boolean whether the window manager is in "showing
                             desktop" mode or not.
    :type showing_desktop:   CARDINAL/32
    :rtype:                  xcb.VoidCookie
    """
    return revent(root, '_NET_SHOWING_DESKTOP', showing_desktop)

def request_showing_desktop_checked(showing_desktop):
    return revent_checked(root, '_NET_SHOWING_DESKTOP', showing_desktop)

# _NET_CLOSE_WINDOW

def request_close_window(window, timestamp=xproto.Time.CurrentTime,
                         source=1):
    """
    Sends event to root window to close a window.

    :param window:      A window identifier.
    :param source:      The source indication.
    :type timestamp:    Milliseconds
    :rtype:             xcb.VoidCookie
    """
    return revent(window, '_NET_CLOSE_WINDOW', timestamp, source)

def request_close_window_checked(window, timestamp=xproto.Time.CurrentTime,
                                 source=1):
    return revent_checked(window, '_NET_CLOSE_WINDOW', timestamp, source)

# _NET_MOVERESIZE_WINDOW

def request_moveresize_window(window, x=None, y=None, width=None,
                              height=None,
                              gravity=xproto.Gravity.BitForget, source=1):
    """
    Sends event to root window to move/resize a window.

    :param window:      A window identifier.
    :param x:           x coordinate
    :param y:           y coordinate
    :param width:       Width
    :param height:      Height
    :param gravity:     Should be one of NorthWest, North, NorthEast, West,
                        Center, East, SouthWest, South, SouthEast, and Static.
                        If set to 0, the window manager should use the default
                        gravity for the window.
    :param source:      The source indication.
    :rtype:             xcb.VoidCookie
    """
    flags = gravity
    flags |= source << 12
    if x is not None:
        flags |= 1 << 8
    if y is not None:
        flags |= 1 << 9
    if width is not None:
        flags |= 1 << 10
    if height is not None:
        flags |= 1 << 11

    return revent(window, '_NET_MOVERESIZE_WINDOW',
                  flags, x or 0, y or 0, width or 0, height or 0)

def request_moveresize_window_checked(window, x=None, y=None, width=None,
                                      height=None,
                                      gravity=xproto.Gravity.BitForget,
                                      source=1):
    flags = gravity
    flags |= source << 12
    if x is not None:
        flags |= 1 << 8
    if y is not None:
        flags |= 1 << 9
    if width is not None:
        flags |= 1 << 10
    if height is not None:
        flags |= 1 << 11

    return revent_checked(window, '_NET_MOVERESIZE_WINDOW',
                          flags, x, y, width, height)

# _NET_WM_MOVERESIZE

def request_wm_moveresize(window, direction, x_root=0, y_root=0,
                          button=0, source=1):
    """
    Sends event to root window to initiate window movement or resizing.

    :param window:      A window identifier.
    :param direction:   Whether it is moving or resizing, and if resizing,
                        the direction. Can be one of the following flags:

                        _NET_WM_MOVERESIZE_SIZE_TOPLEFT         = 0

                        _NET_WM_MOVERESIZE_SIZE_TOP             = 1

                        _NET_WM_MOVERESIZE_SIZE_TOPRIGHT        = 2

                        _NET_WM_MOVERESIZE_SIZE_RIGHT           = 3

                        _NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT     = 4

                        _NET_WM_MOVERESIZE_SIZE_BOTTOM          = 5

                        _NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT      = 6

                        _NET_WM_MOVERESIZE_SIZE_LEFT            = 7

                        _NET_WM_MOVERESIZE_MOVE                 = 8

                        _NET_WM_MOVERESIZE_SIZE_KEYBOARD        = 9

                        _NET_WM_MOVERESIZE_MOVE_KEYBOARD        = 10

                        _NET_WM_MOVERESIZE_CANCEL               = 11
    :param x_root:      x coordinate of the pointer.
    :param y_root:      y coordinate of the pointer.
    :param button:      Which button was pressed, if a mouse button was used
                        to initiate this request.
    :param source:      The source indication.
    :rtype:             xcb.VoidCookie
    """
    return revent(window, '_NET_WM_MOVERESIZE',
                  x_root, y_root, direction, button, source)

def request_wm_moveresize_checked(window, direction, x_root=0, y_root=0,
                                  button=0, source=1):
    return revent_checked(window, '_NET_WM_MOVERESIZE',
                          x_root, y_root, direction, button, source)

# _NET_RESTACK_WINDOW

def request_restack_window(window, stack_mode=xproto.StackMode.Above,
                           sibling=0, source=1):
    """
    Sends event to root window to restack a window.

    :param window:      A window identifier.
    :param stack_mode:  Stacking mode of window. Can be one of the following
                        flags: Above, Below, TopIf, BottomIf, Opposite
    :param sibling:     A sibling window identifier.
    :param source:      The source indication.
    :rtype:             xcb.VoidCookie
    """
    return revent(window, '_NET_RESTACK_WINDOW', source, sibling, stack_mode)

def request_restack_window_checked(window,
                                   stack_mode=xproto.StackMode.Above,
                                   sibling=0, source=2):
    return revent_checked(window, '_NET_RESTACK_WINDOW',
                          source, sibling, stack_mode)

# _NET_REQUEST_FRAME_EXTENTS

def request_request_frame_extents(window):
    """
    Sends event to root window ask the WM to estimate the frame extents.

    :param window:      A window identifier.
    :rtype:             xcb.VoidCookie
    """
    return revent(window, '_NET_REQUEST_FRAME_EXTENTS')

def request_request_frame_extents_checked(window):
    return revent_checked(window, '_NET_REQUEST_FRAME_EXTENTS')

# _NET_WM_NAME

def get_wm_name(window):
    """
    Get the title of a window.

    :param window:  A window identifier.
    :return:        The window's title.
    :rtype:         util.PropertyCookie (UTF8_STRING)
    """
    return util.PropertyCookie(util.get_property(window, '_NET_WM_NAME'))

def get_wm_name_unchecked(window):
    return util.PropertyCookie(util.get_property_unchecked(window,
                                                           '_NET_WM_NAME'))

def set_wm_name(window, wm_name):
    """
    Sets the title of a window.

    :param window:  A window identifier.
    :param wm_name: The title of the window.
    :rtype:         xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_NAME'),
                                 atom('UTF8_STRING'), 8, len(wm_name), wm_name)

def set_wm_name_checked(window, wm_name):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_NAME'),
                                        atom('UTF8_STRING'), 8, len(wm_name),
                                        wm_name)

# _NET_WM_VISIBLE_NAME

def get_wm_visible_name(window):
    """
    Get the visible title of a window.

    :param window:  A window identifier.
    :return:        The window's visible title.
    :rtype:         util.PropertyCookie (UTF8_STRING)
    """
    return util.PropertyCookie(util.get_property(window,
                                                 '_NET_WM_VISIBLE_NAME'))

def get_wm_visible_name_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_VISIBLE_NAME')
    return util.PropertyCookie(cook)

def set_wm_visible_name(window, wm_name):
    """
    Sets the visible title of a window.

    :param window:  A window identifier.
    :param wm_name: The title of the window.
    :rtype:         xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_VISIBLE_NAME'),
                                 atom('UTF8_STRING'), 8, len(wm_name), wm_name)

def set_wm_visible_name_checked(window, wm_name):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_VISIBLE_NAME'),
                                        atom('UTF8_STRING'), 8, len(wm_name),
                                        wm_name)

# _NET_WM_ICON_NAME

def get_wm_icon_name(window):
    """
    Get the icon name of a window.

    :param window:  A window identifier.
    :return:        The window's icon name.
    :rtype:         util.PropertyCookie (UTF8_STRING)
    """
    return util.PropertyCookie(util.get_property(window, '_NET_WM_ICON_NAME'))

def get_wm_icon_name_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_ICON_NAME')
    return util.PropertyCookie(cook)

def set_wm_icon_name(window, icon_name):
    """
    Sets the icon name of a window.

    :param window:      A window identifier.
    :param icon_name:   The icon name of the window.
    :rtype:             xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_ICON_NAME'), atom('UTF8_STRING'),
                                 8, len(icon_name), icon_name)

def set_wm_icon_name_checked(window, icon_name):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_ICON_NAME'),
                                        atom('UTF8_STRING'), 8, len(icon_name),
                                        icon_name)

# _NET_WM_VISIBLE_ICON_NAME

def get_wm_visible_icon_name(window):
    """
    Get the visible icon name of a window.

    :param window:  A window identifier.
    :return:        The window's visible icon name.
    :rtype:         util.PropertyCookie (UTF8_STRING)
    """
    return util.PropertyCookie(util.get_property(window,
                                                 '_NET_WM_VISIBLE_ICON_NAME'))

def get_wm_visible_icon_name_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_VISIBLE_ICON_NAME')
    return util.PropertyCookie(cook)

def set_wm_visible_icon_name(window, icon_name):
    """
    Sets the visible icon name of a window.

    :param window:      A window identifier.
    :param icon_name:   The icon name of the window.
    :rtype:             xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_VISIBLE_ICON_NAME'),
                                 atom('UTF8_STRING'), 8, len(icon_name),
                                 icon_name)

def set_wm_visible_icon_name_checked(window, icon_name):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_VISIBLE_ICON_NAME'),
                                        atom('UTF8_STRING'), 8, len(icon_name),
                                        icon_name)

# _NET_WM_WINDOW_OPACITY

class OpacityCookieSingle(util.PropertyCookieSingle):
    def reply(self):
        v = util.PropertyCookieSingle.reply(self)

        if not v:
            return None

        return float(v) / float(0xffffffff)

def get_wm_window_opacity(window):
    """
    Get the opacity of the current window.

    N.B. If your window manager uses decorations, you'll typically want to pass
    your client's *parent* window to this function.

    :param window:  A window identifier.
    :return:        An opacity percentage between 0 and 1 (inclusive)
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    return OpacityCookieSingle(util.get_property(window,
                                                 '_NET_WM_WINDOW_OPACITY'))

def get_wm_window_opacity_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_WINDOW_OPACITY')
    return OpacityCookieSingle(cook)

def set_wm_window_opacity(window, opacity):
    """
    Sets the opacity of the current window.

    N.B. If your window manager uses decorations, you'll typically want to pass
    your client's *parent* window to this function.

    :param window:  A window identifier.
    :param opacity: A float between 0 and 1 inclusive.

                    0 is completely transparent and 1 is completely opaque.
    :type opacity:  CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    assert 0 <= opacity <= 1
    packed = struct.pack('I', int(opacity * 0xffffffff))
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_WINDOW_OPACITY'),
                                 CARDINAL, 32, 1, packed)

def set_wm_window_opacity_checked(window, opacity):
    assert 0 <= opacity <= 1
    packed = struct.pack('I', int(opacity * 0xffffffff))
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_WINDOW_OPACITY'),
                                        CARDINAL, 32, 1, packed)

# _NET_WM_DESKTOP

def get_wm_desktop(window):
    """
    Get the desktop index of the window.

    :param window:  A window identifier.
    :return:        The window's virtual desktop index.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    return util.PropertyCookieSingle(util.get_property(window,
                                                       '_NET_WM_DESKTOP'))

def get_wm_desktop_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_DESKTOP')
    return util.PropertyCookieSingle(cook)

def set_wm_desktop(window, desktop):
    """
    Sets the desktop index of the window.

    :param window:  A window identifier.
    :param desktop: A desktop index.
    :type desktop:  CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_DESKTOP'),
                                 CARDINAL, 32, 1, [desktop])

def set_wm_desktop_checked(window, desktop):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_DESKTOP'),
                                        CARDINAL, 32, 1, [desktop])

def request_wm_desktop(window, desktop, source=1):
    """
    Sends an event to root window to change the desktop of the window.

    :param window:  A window identifier.
    :param desktop: A desktop index.
    :type desktop:  CARDINAL/32
    :param source:  The source indication.
    :rtype:         xcb.VoidCookie
    """
    return revent(window, '_NET_WM_DESKTOP', desktop, source)

def request_wm_desktop_checked(window, desktop, source=1):
    return revent_checked(window, '_NET_WM_DESKTOP', desktop, source)

# _NET_WM_WINDOW_TYPE

def get_wm_window_type(window):
    """
    Get a list of atoms representing the type of the window.

    :param window:  A window identifier.
    :return:        A list of atoms corresponding to this window's type.
    :rtype:         util.PropertyCookie (ATOM[]/32)
    """
    return util.PropertyCookie(util.get_property(window, '_NET_WM_WINDOW_TYPE'))

def get_wm_window_type_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_WINDOW_TYPE')
    return util.PropertyCookie(cook)

def set_wm_window_type(window, types):
    """
    Sets the list of atoms representing this window's type.

    :param window:  A window identifier.
    :param types:   A list of window type atoms.
    :type types:    ATOM[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(types), *types)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_WINDOW_TYPE'),
                                 ATOM, 32, len(types), packed)

def set_wm_window_type_checked(window, types):
    packed = struct.pack('I' * len(types), *types)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_WINDOW_TYPE'),
                                        ATOM, 32, len(types), packed)

# _NET_WM_STATE

def get_wm_state(window):
    """
    Get a list of atoms representing the state of the window.

    :param window:  A window identifier.
    :return:        A list of atoms corresponding to this window's state.
    :rtype:         util.PropertyCookie (ATOM[]/32)
    """
    return util.PropertyCookie(util.get_property(window, '_NET_WM_STATE'))

def get_wm_state_unchecked(window):
    return util.PropertyCookie(util.get_property_unchecked(window,
                                                           '_NET_WM_STATE'))

def set_wm_state(window, states):
    """
    Sets the list of atoms representing this window's state.

    :param window:  A window identifier.
    :param states:  A list of window state atoms.
    :type states:   ATOM[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(states), *states)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_STATE'), ATOM, 32, len(states),
                                 packed)

def set_wm_state_checked(window, states):
    packed = struct.pack('I' * len(states), *states)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_STATE'),
                                        ATOM, 32, len(states), packed)

def request_wm_state(window, action, first, second=0, source=1):
    """
    Sends an event to root window to change the state of the window.

    :param window:  A window identifier.
    :param action:  The kind of state action to perform. I tshould be one
                    of the following flags:

                    _NET_WM_STATE_REMOVE    = 0
                    _NET_WM_STATE_ADD       = 1
                    _NET_WM_STATE_TOGGLE    = 2
    :param first:   The first property to be changed.
    :param second:  The second property to be changed (should be 0 if only
                    one property is being changed).
    :param source:  The source indication.
    :rtype:         xcb.VoidCookie
    """
    return revent(window, '_NET_WM_STATE', action, first, second, source)

def request_wm_state_checked(window, action, first, second=0, source=1):
    return revent_checked(window, '_NET_WM_STATE',
                          action, first, second, source)

# _NET_WM_ALLOWED_ACTIONS

def get_wm_allowed_actions(window):
    """
    Get a list of atoms representing the WM supported actions on a window.

    :param window:  A window identifier.
    :return:        A list of atoms corresponding to this window's supported
                    actions through the window manager.
    :rtype:         util.PropertyCookie (ATOM[]/32)
    """
    return util.PropertyCookie(util.get_property(window,
                                                 '_NET_WM_ALLOWED_ACTIONS'))

def get_wm_allowed_actions_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_ALLOWED_ACTIONS')
    return util.PropertyCookie(cook)

def set_wm_allowed_actions(window, actions):
    """
    Sets the list of atoms representing the WM supported actions on a window.

    :param window:  A window identifier.
    :param actions:  A list of allowable action atoms.
    :type actions:   ATOM[]/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I' * len(actions), *actions)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_ALLOWED_ACTIONS'),
                                 ATOM, 32, len(actions), packed)

def set_wm_allowed_actions_checked(window, actions):
    packed = struct.pack('I' * len(actions), *actions)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_ALLOWED_ACTIONS'),
                                        ATOM, 32, len(actions), packed)

# _NET_WM_STRUT

class StrutCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'left': v[0],
            'right': v[1],
            'top': v[2],
            'bottom': v[3]
        }

def get_wm_strut(window):
    """
    Returns the struts for a window.

    :param window:  A window identifier.
    :return:        A strut dictionary.

                    Keys: left, right, top, bottom
    :rtype:         StrutCookie (CARDINAL[4]/32)
    """
    return StrutCookie(util.get_property(window, '_NET_WM_STRUT'))

def get_wm_strut_unchecked(window):
    return StrutCookie(util.get_property_unchecked(window, '_NET_WM_STRUT'))

def set_wm_strut(window, left, right, top, bottom):
    """
    Sets the struts for a window.

    :param window:  A window identifier.
    :param left:    Width of area at left side of screen.
    :type left:     CARDINAL/32
    :param right:   Width of area at right side of screen.
    :type right:    CARDINAL/32
    :param top:     Height of area at top side of screen.
    :type top:      CARDINAL/32
    :param bottom:  Height of area at bottom side of screen.
    :type bottom:   CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('IIII', left, right, top, bottom)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_STRUT'), CARDINAL, 32, 4, packed)

def set_wm_strut_checked(window, left, right, top, bottom):
    packed = struct.pack('IIII', left, right, top, bottom)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_STRUT'),
                                        CARDINAL, 32, 4, packed)

# _NET_WM_STRUT_PARTIAL

class StrutPartialCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'left': v[0],
            'right': v[1],
            'top': v[2],
            'bottom': v[3],
            'left_start_y': v[4],
            'left_end_y': v[5],
            'right_start_y': v[6],
            'right_end_y': v[7],
            'top_start_x': v[8],
            'top_end_x': v[9],
            'bottom_start_x': v[10],
            'bottom_end_x': v[11]
        }

def get_wm_strut_partial(window):
    """
    Returns the partial struts for a window.

    :param window:  A window identifier.
    :return:        A strut dictionary.

                    Keys: left, right, top, bottom, left_start_y, left_end_y,
                    right_start_y, right_end_y, top_start_x, top_end_x,
                    bottom_start_x, bottom_end_x
    :rtype:         StrutPartialCookie (CARDINAL[12]/32)
    """
    return StrutPartialCookie(util.get_property(window,
                                                '_NET_WM_STRUT_PARTIAL'))

def get_wm_strut_partial_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_STRUT_PARTIAL')
    return StrutPartialCookie(cook)

def set_wm_strut_partial(window, left, right, top, bottom, left_start_y,
                         left_end_y, right_start_y, right_end_y, top_start_x,
                         top_end_x, bottom_start_x, bottom_end_x):
    """
    Sets the partial struts for a window.

    :param window:          A window identifier.
    :param left:            Width of area at left side of screen.
    :type left:             CARDINAL/32
    :param right:           Width of area at right side of screen.
    :type right:            CARDINAL/32
    :param top:             Height of area at top side of screen.
    :type top:              CARDINAL/32
    :param bottom:          Height of area at bottom side of screen.
    :type bottom:           CARDINAL/32
    :param left_start_y:
    :param left_end_y:
    :param right_start_y:
    :param right_end_y:
    :param top_start_x:
    :param top_end_x:
    :param bottom_start_x:
    :param bottom_end_x:
    :rtype:                 xcb.VoidCookie
    """
    packed = struct.pack('I' * 12, left, right, top, bottom, left_start_y,
                         left_end_y, right_start_y, right_end_y, top_start_x,
                         top_end_x, bottom_start_x, bottom_end_x)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_STRUT_PARTIAL'),
                                 CARDINAL, 32, 12, packed)

def set_wm_strut_partial_checked(window, left, right, top, bottom,
                                 left_start_y, left_end_y, right_start_y,
                                 right_end_y, top_start_x, top_end_x,
                                 bottom_start_x, bottom_end_x):
    packed = struct.pack('I' * 12, left, right, top, bottom, left_start_y,
                         left_end_y, right_start_y, right_end_y, top_start_x,
                         top_end_x, bottom_start_x, bottom_end_x)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_STRUT_PARTIAL'),
                                        CARDINAL, 32, 12, packed)

# _NET_WM_ICON_GEOMETRY

class IconGeometryCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'x': v[0],
            'y': v[1],
            'width': v[2],
            'height': v[3]
        }

def get_wm_icon_geometry(window):
    """
    Returns the icon geometry for a window.

    :param window:  A window identifier.
    :return:        An icon geometry dictionary.

                    Keys: x, y, width, height
    :rtype:         IconGeometryCookie (CARDINAL[4]/32)
    """
    return IconGeometryCookie(util.get_property(window,
                                                '_NET_WM_ICON_GEOMETRY'))

def get_wm_icon_geometry_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_ICON_GEOMETRY')
    return IconGeometryCookie(cook)

def set_wm_icon_geometry(window, x, y, width, height):
    """
    Sets the icon geometry for a window.

    :param window:  A window identifier.
    :param x:       x coordinate of icon area.
    :type x:        CARDINAL/32
    :param y:       y coordinate of icon area.
    :type y:        CARDINAL/32
    :param width:   Width of icon area.
    :type width:    CARDINAL/32
    :param height:  Height of icon area.
    :type height:   CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('IIII', x, y, width, height)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_ICON_GEOMETRY'),
                                 CARDINAL, 32, 4, packed)

def set_wm_icon_geometry_checked(window, x, y, width, height):
    packed = struct.pack('IIII', x, y, width, height)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_ICON_GEOMETRY'),
                                        CARDINAL, 32, 4, packed)

# _NET_WM_ICON

class IconCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        ret = []
        start = 0
        while start < len(v):
            w, h = v[start], v[start + 1]
            upto = w * h

            ret.append({
                'width': w,
                'height': h,
                'data': v[start + 2:start + upto + 2]
            })

            start = start + upto + 2

        return ret

def get_wm_icon(window):
    """
    Returns an array of possible icons for a window.

    :param window:  A window identifier.
    :return:        A list of icon dictionaries.

                    Keys: width, height, data
    :rtype:         IconCookie (CARDINAL[][2+n]/32)
    """
    return IconCookie(util.get_property(window, '_NET_WM_ICON'))

def get_wm_icon_unchecked(window):
    return IconCookie(util.get_property_unchecked(window, '_NET_WM_ICON'))

def set_wm_icon(window, icons):
    """
    Sets the array of possible icons for a window.

    :param window:  A window identifier.
    :param icons:   A list icon dictionaries. Each dictionary should have
                    the following keys: width, height and data.
    :type icons:    CARDINAL[][2+n]/32
    :rtype:         xcb.VoidCookie
    """
    flatten = []
    for icon in icons:
        flatten.append(icon['width'])
        flatten.append(icon['height'])
        for argb in icon['data']:
            flatten.append(argb)
    packed = struct.pack('I' * len(flatten), *flatten)

    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_ICON'),
                                 CARDINAL, 32, len(flatten), packed)

def set_wm_icon_checked(window, icons):
    flatten = []
    for icon in icons:
        flatten.append(icon['width'])
        flatten.append(icon['height'])
        for argb in icon['data']:
            flatten.append(argb)
    packed = struct.pack('I' * len(flatten), *flatten)

    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_ICON'),
                                        CARDINAL, 32, len(flatten), packed)

# _NET_WM_PID

def get_wm_pid(window):
    """
    Get the process ID of the client owning a window.

    :param window:  A window identifier.
    :return:        The window's client's process ID.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    return util.PropertyCookieSingle(util.get_property(window, '_NET_WM_PID'))

def get_wm_pid_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_PID')
    return util.PropertyCookieSingle(cook)

def set_wm_pid(window, pid):
    """
    Sets the process ID of the client owning a window.

    :param window:  A window identifier.
    :param pid:     A process ID.
    :type pid:      CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('I', pid)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_PID'), CARDINAL, 32, 1,
                                 packed)

def set_wm_pid_checked(window, pid):
    packed = struct.pack('I', pid)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_PID'), CARDINAL, 32, 1,
                                        packed)

# _NET_WM_HANDLED_ICONS

def get_wm_handled_icons(window):
    """
    Gets the "handled icons" property.

    :param window:  A window identifier.
    :return:        Whether this property is set or not.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    return util.PropertyCookieSingle(util.get_property(window,
                                                       '_NET_WM_HANDLED_ICONS'))

def get_wm_handled_icons_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_HANDLED_ICONS')
    return util.PropertyCookieSingle(cook)

def set_wm_handled_icons(window):
    """
    Sets the "handled icons" property.

    :param window:  A window identifier.
    :return:        Whether this property is set or not.
    :rtype:         xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_HANDLED_ICONS'), CARDINAL, 32, 1,
                                 [1])

def set_wm_handled_icons_checked(window):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_HANDLED_ICONS'),
                                        CARDINAL, 32, 1, [1])

# _NET_WM_USER_TIME

def get_wm_user_time(window):
    """
    Get the time at which the last user activity occurred on a window.

    :param window:  A window identifier.
    :return:        The XServer time when user activity last occurred.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    return util.PropertyCookieSingle(util.get_property(window,
                                                       '_NET_WM_USER_TIME'))

def get_wm_user_time_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_USER_TIME')
    return util.PropertyCookieSingle(cook)

def set_wm_user_time(window, user_time):
    """
    Sets the time that user activity last occurred on this window.

    :param window:      A window identifier.
    :param user_time:   Last user activity time.
    :type user_time:    CARDINAL/32
    :rtype:             xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_USER_TIME'), CARDINAL, 32, 1,
                                 [user_time])

def set_wm_user_time_checked(window, user_time):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_USER_TIME'),
                                        CARDINAL, 32, 1, [user_time])

# _NET_WM_USER_TIME_WINDOW

def get_wm_user_time_window(window):
    """
    Gets the window that sets the _NET_WM_USER_TIME property.

    :param window:  A window identifier.
    :return:        Window identifier that sets the _NET_WM_USER_TIME property.
    :rtype:         util.PropertyCookieSingle (WINDOW/32)
    """
    cook = util.get_property(window, '_NET_WM_USER_TIME_WINDOW')
    return util.PropertyCookieSingle(cook)

def get_wm_user_time_window_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_USER_TIME_WINDOW')
    return util.PropertyCookieSingle(cook)

def set_wm_user_time_window(window, time_win):
    """
    Sets the window identifier that sets the _NET_WM_USER_TIME property.

    :param window:      A window identifier.
    :param time_win:    Window ID that sets the _NET_WM_USER_TIME property.
    :type time_win:     WINDOW/32
    :rtype:             xcb.VoidCookie
    """
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_USER_TIME_WINDOW'),
                                 WINDOW, 32, 1, [time_win])

def set_wm_user_time_window_checked(window, time_win):
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_USER_TIME_WINDOW'),
                                        WINDOW, 32, 1, [time_win])

# _NET_FRAME_EXTENTS

class FrameExtentsCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'left': v[0],
            'right': v[1],
            'top': v[2],
            'bottom': v[3]
        }

def get_frame_extents(window):
    """
    Returns the frame extents for a window.

    :param window:  A window identifier.
    :return:        A frame extents dictionary.

                    Keys: left, right, top, bottom
    :rtype:         FrameExtentsCookie (CARDINAL[4]/32)
    """
    return FrameExtentsCookie(util.get_property(window, '_NET_FRAME_EXTENTS'))

def get_frame_extents_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_FRAME_EXTENTS')
    return FrameExtentsCookie(cook)

def set_frame_extents(window, left, right, top, bottom):
    """
    Sets the frame extents for a window.

    :param window:  A window identifier.
    :param left:    Width of left border.
    :type left:     CARDINAL/32
    :param right:   Width of right border.
    :type right:    CARDINAL/32
    :param top:     Height of top border.
    :type top:      CARDINAL/32
    :param bottom:  Height of bottom border.
    :type bottom:   CARDINAL/32
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('IIII', left, right, top, bottom)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_FRAME_EXTENTS'), CARDINAL, 32, 4,
                                 packed)

def set_frame_extents_checked(window, left, right, top, bottom):
    packed = struct.pack('IIII', left, right, top, bottom)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_FRAME_EXTENTS'),
                                        CARDINAL, 32, 4, packed)

# _NET_WM_PING

def request_wm_ping(window, response=False,
                    timestamp=xproto.Time.CurrentTime):
    """
    Sends an event to root window to ping a window or respond to a ping.

    :param window:      A window identifier.
    :param response:    Whether this is a response to a ping request or not.
    :type timestamp:    Milliseconds
    :rtype:             xcb.VoidCookie
    """
    return revent(window if not response else root(c),
                  'WM_PROTOCOLS', atom('_NET_WM_PING'), timestamp, window)

def request_wm_ping_checked(window, response=False,
                            timestamp=xproto.Time.CurrentTime):
    return revent_checked(window if not response else root(c), 'WM_PROTOCOLS',
                          atom('_NET_WM_PING'), timestamp, window)

# _NET_WM_SYNC_REQUEST

def request_wm_sync_request(window, req_num,
                            timestamp=xproto.Time.CurrentTime):
    """
    Sends an event to root window to sync with a client.

    :param window:      A window identifier.
    :param req_num:     The XSync request number.
    :type timestamp:    Milliseconds
    :rtype:             xcb.VoidCookie
    """
    high = req_num >> 32
    low = (high << 32) ^ req_num

    return revent(window, 'WM_PROTOCOLS',
                  atom('_NET_WM_SYNC_REQUEST'), timestamp, low, high)

def request_wm_sync_request_checked(window, req_num,
                                    timestamp=xproto.Time.CurrentTime):
    high = req_num >> 32
    low = (high << 32) ^ req_num

    return revent_checked(window, 'WM_PROTOCOLS',
                          atom('_NET_WM_SYNC_REQUEST'), timestamp, low, high)

# _NET_WM_SYNC_REQUEST_COUNTER

def get_wm_sync_request_counter(window):
    """
    Gets XSync counter for this client.

    :param window:  A window identifier.
    :return:        An XSync XID.
    :rtype:         util.PropertyCookieSingle (CARDINAL/32)
    """
    cook = util.get_property(window, '_NET_WM_SYNC_REQUEST_COUNTER')
    return util.PropertyCookieSingle(cook)

def get_wm_sync_request_counter_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_SYNC_REQUEST_COUNTER')
    return util.PropertyCookieSingle(cook)

def set_wm_sync_request_counter(window, counter):
    """
    Sets the XSync counter for this client.

    :param window:      A window identifier.
    :param counter:     An XSync XID.
    :type counter:      CARDINAL
    :rtype:             xcb.VoidCookie
    """
    packed = struct.pack('I', counter)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_SYNC_REQUEST_COUNTER'),
                                 CARDINAL, 32, 1, packed)

def set_wm_sync_request_counter_checked(window, counter):
    packed = struct.pack('I', counter)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_SYNC_REQUEST_COUNTER'),
                                        CARDINAL, 32, 1, packed)

# _NET_WM_FULLSCREEN_MONITORS

class FullscreenMonitorsCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'top': v[0],
            'bottom': v[1],
            'left': v[2],
            'right': v[3]
        }

def get_wm_fullscreen_monitors(window):
    """
    Get list of monitor edges for this window.

    :param window:  A window identifier.
    :return:        The window's monitor edges.
    :rtype:         FullscreenMonitorsCookie (CARDINAL[4]/32)
    """
    return FullscreenMonitorsCookie(
        util.get_property(window, '_NET_WM_FULLSCREEN_MONITORS'))

def get_wm_fullscreen_monitors_unchecked(window):
    cook = util.get_property_unchecked(window, '_NET_WM_FULLSCREEN_MONITORS')
    return FullscreenMonitorsCookie(cook)

def set_wm_fullscreen_monitors(window, top, bottom, left, right):
    """
    Sets list of monitor edges for this window.

    :param window:  A window identifier.
    :param top:     The monitor whose top edge defines the top edge of window.
    :param bottom:  The monitor whose bottom edge defines the bottom edge of
                    window.
    :param left:    The monitor whose left edge defines the left edge of
                    window.
    :param right:   The monitor whose right edge defines the right edge of
                    window.
    :rtype:         xcb.VoidCookie
    """
    packed = struct.pack('IIII', top, bottom, left, right)
    return c.core.ChangeProperty(xproto.PropMode.Replace, window,
                                 atom('_NET_WM_FULLSCREEN_MONITORS'),
                                 CARDINAL, 32, 4, packed)

def set_wm_fullscreen_monitors_checked(window, top, bottom, left, right):
    packed = struct.pack('IIII', top, bottom, left, right)
    return c.core.ChangePropertyChecked(xproto.PropMode.Replace, window,
                                        atom('_NET_WM_FULLSCREEN_MONITORS'),
                                        CARDINAL, 32, 4, packed)

def request_wm_fullscreen_monitors(window, top, bottom, left, right,
                                   source=1):
    """
    Sends an event to root window to change monitor edges for this window.

    :param window:  A window identifier.
    :param top:     The monitor whose top edge defines the top edge of window.
    :param bottom:  The monitor whose bottom edge defines the bottom edge of
                    window.
    :param left:    The monitor whose left edge defines the left edge of
                    window.
    :param right:   The monitor whose right edge defines the right edge of
                    window.
    :param source:  The source indication.
    :rtype:         xcb.VoidCookie
    """
    return revent(window, '_NET_WM_FULLSCREEN_MONITORS',
                  top, bottom, left, right, source)

def request_wm_fullscreen_monitors_checked(window, top, bottom, left, right,
                                           source=1):
    return revent_checked(window, '_NET_WM_FULLSCREEN_MONITORS',
                          top, bottom, left, right, source)




# Makes all inactive windows transparent and keeps the active window opaque.
# This should be run in the background as a daemon like so:
# python2 inactive-window-transparent.py

# Range of values: 0 <= opacity <= 1
# where 1 is fully opaque and 0 is completely invisible
opacity = 0.8

import xpybutil
import xpybutil.event as event
import xpybutil.ewmh as ewmh
import xpybutil.util as util
import xpybutil.window as window

def update_window_opacity():
    activewin = ewmh.get_active_window().reply()
    if not activewin:
        return

    for client in clients:
        ewmh.set_wm_window_opacity(util.get_parent_window(client), 
                                   1 if client == activewin else opacity)

    xpybutil.conn.flush()

def client_is_normal(client):
    wtype = ewmh.get_wm_window_type(client).reply()
    if not wtype or wtype[0] == util.get_atom('_NET_WM_WINDOW_TYPE_NORMAL'):
        return True
    return False

def cb_property_notify(e):
    global clients

    aname = util.get_atom_name(e.atom)

    if aname == '_NET_ACTIVE_WINDOW':
        update_window_opacity();
    elif aname == '_NET_CLIENT_LIST':
        clients = filter(client_is_normal, ewmh.get_client_list().reply())

clients = filter(client_is_normal, ewmh.get_client_list().reply())
update_window_opacity()

window.listen(xpybutil.root, 'PropertyChange')
event.connect('PropertyNotify', xpybutil.root, cb_property_notify)

event.main()




"""
A set of functions to faciliate binding to mouse button presses.

This is basically the companion module of 'keybind.py'. The good news is that
we don't have to deal with the complexity of X's keysym table. The bad news is
that mice binding and dragging comes with its own hairball of complexity.

TODO: Mouse bindings that hook into the event dipatcher like the keybind
      module. Also, functions that abstract the process of using the mouse
      to drag something.
"""
from collections import defaultdict

from xpybutil.compat import xproto

from xpybutil import conn, root

__mousebinds = defaultdict(list)
__mousegrabs = defaultdict(int) # Mouse grab key -> number of grabs

EM = xproto.EventMask
GM = xproto.GrabMode
TRIVIAL_MODS = [
    0,
    xproto.ModMask.Lock,
    xproto.ModMask._2,
    xproto.ModMask.Lock | xproto.ModMask._2
]

def parse_buttonstring(button_string):
    """
    A utility function to turn strings like 'Mod1-Shift-3' into a pair
    corresponding to its modifiers and button number.

    :param button_string: String starting with zero or more modifiers followed
                          by exactly one button number.

                          Available modifiers: Control, Mod1, Mod2, Mod3, Mod4,
                          Mod5, Shift, Lock
    :type button_string: str
    :return: Tuple of modifier mask and button number
    :rtype: (mask, int)
    """
    mods, button = 0, None
    for part in button_string.split('-'):
        if hasattr(xproto.KeyButMask, part):
            mods |= getattr(xproto.KeyButMask, part)
        else:
            button = int(part)

    return mods, button

def grab_pointer(grab_win, confine, cursor):
    """
    This will grab the pointer. The effect is that further pointer events will
    *only* be sent to the grabbing client (i.e., ``grab_win``).

    :param grab_win: A window identifier to report pointer events to.
    :type grab_win: int
    :param confine: Either a window identifier to confine the pointer to, or
                    None if no confinement is desired.
    :type confine: int
    :param cursor: A cursor identifier. (See the ``cursor`` module.)
    :rtype: xcb.xproto.GrabStatus
    """
    mask = EM.PointerMotion | EM.ButtonRelease | EM.ButtonPress
    return conn.core.GrabPointer(False, grab_win, mask, GM.Async, GM.Async,
                                 confine, cursor,
                                 xproto.Time.CurrentTime).reply()

def ungrab_pointer():
    """
    This will release a grab initiated by ``grab_pointer``.

    :rtype: void
    """
    conn.core.UngrabPointerChecked(xproto.Time.CurrentTime).check()

def grab_button(wid, modifiers, button, propagate=False):
    """
    Grabs a mouse button for a particular window and a modifiers/key value.
    If the grab was successful, return True. Otherwise, return False.
    If your client is grabbing buttons, it is useful to notify the user if a
    button wasn't grabbed. Mouse shortcuts not responding is disorienting!

    Also, this function will grab several buttons based on varying modifiers.
    Namely, this accounts for all of the "trivial" modifiers that may have
    an effect on X events, but probably shouldn't effect button grabbing. (i.e.,
    whether num lock or caps lock is on.)

    If 'propagate' is True, then no further events can be processed until the
    grabbing client allows them to be. (Which is done via AllowEvents. Thus,
    if propagate is True, you *must* make some call to AllowEvents at some
    point, or else your client will lock.)

    N.B. You should probably be using 'bind_mouse' or 'bind_global_mouse'
    instead.

    :param wid: A window identifier.
    :type wid: int
    :param modifiers: A modifier mask.
    :type modifiers: int
    :param button: A button number.
    :type button: int
    :param propagate: Whether to grab synchronously (True) or
                      asynchronously (False). When in doubt, keep it set
                      to False.
    :type propagate: bool
    :rtype: bool
    """
    mask = EM.ButtonPress | EM.ButtonRelease | EM.ButtonMotion

    try:
        for mod in TRIVIAL_MODS:
            conn.core.GrabButtonChecked(True, wid, mask,
                                        GM.Sync if propagate else GM.Async,
                                        GM.Async, 0, 0,
                                        button, modifiers | mod).check()

        return True
    except xproto.BadAccess:
        return False

def ungrab_button(wid, modifiers, button):
    """
    Ungrabs a button that was grabbed by ``grab_button``. Similarly, it
    will return True on success and False on failure.

    When ungrabbing a button, the parameters to this function should be
    *precisely* the same as the parameters to ``grab_button``.

    :param wid: A window identifier.
    :type wid: int
    :param modifiers: A modifier mask.
    :type modifiers: int
    :param button: A button number.
    :type button: int
    :rtype: bool
    """
    try:
        for mod in TRIVIAL_MODS:
            conn.core.UngrabButtonChecked(button, wid, modifiers | mod).check()

        return True
    except xproto.BadAccess:
        return False

def bind_global_mouse(event_type, key_string, cb):
    """
    An alias for ``bind_mouse(event_type, ROOT_WINDOW, key_string, cb)``.

    :param event_type: Either 'ButtonPress' or 'ButtonRelease'.
    :type event_type: str
    :param key_string: A string of the form 'Mod1-Control-3'.
                       Namely, a list of zero or more modifiers separated by
                       '-', followed by a single button digit.
    :type key_string: str
    :param cb: A first class function with no parameters.
    :type cb: function
    :return: True if the binding was successful, False otherwise.
    :rtype: bool
    """
    return bind_mouse(event_type, root, key_string, cb)

def bind_mouse(event_type, wid, button_string, cb):
    raise NotImplemented




"""
A couple of functions that support retrieving information about
all active physical heads. This is done through the Xinerama
extension, which implicitly supports RandR and TwinView.

The 'get_physical_mapping' function will produce a list of
monitor indices in a physical ordering (left to right, top to
bottom). These indices can then be used in the list returned by
'get_monitors'.
"""
from xpybutil.compat import xinerama

from xpybutil import conn

ext = None
if conn is not None:
    ext = conn(xinerama.key)

def get_monitors():
    '''
    Returns a list of Xinerama screen rectangles.
    They come in the order that the Xinerama extension specifies.

    :rtype: List of (x, y, w, h) rectangles
    '''
    retval = []
    ms = ext.QueryScreens().reply()
    if ms:
        for m in ms.screen_info:
            retval.append((m.x_org, m.y_org, m.width, m.height))

    return retval

def get_physical_mapping(monitors):
    '''
    Returns a list of Xinerama screen indices in their physical order.
    Where physical order is defined as left-to-right and then top-to-bottom.

    :param monitors:  List of (x, y, w, h) rectangles
    :rtype:           List of Xinerama indices
    '''
    retval = []

    tosort = []
    for i, (x, y, w, h) in enumerate(monitors):
        tosort.append((x, y, i))
    for x, y, xscreen in sorted(tosort):
        retval.append(xscreen)

    return retval




"""
This module has a few utility functions that perform math on rectangles.

For example, finding the area of intersection of two rectangles with
``rect_intersect_area``, or getting the rectangle of a monitor after accounting
for struts with ``monitor_rects``.
"""
from xpybutil.compat import xproto

import xpybutil.ewmh as ewmh
import xpybutil.window as window

def rect_intersect_area(r1, r2):
    """
    Returns the area of the intersection of two rectangles. If the rectangles
    do not intersect, the area returned is 0.

    :param r1: A 4-tuple representing a rectangle:

               (top_left_x, top_left_y, width, height)
    :param r2: A 4-tuple representing a rectangle:

               (top_left_x, top_left_y, width, height)
    :return: Area of intersection of r1 and r2.
    :rtype:  Integer
    """
    x1, y1, w1, h1 = r1
    x2, y2, w2, h2 = r2
    if x2 < x1 + w1 and x2 + w2 > x1 and y2 < y1 + h1 and y2 + h2 > y1:
        iw = min(x1 + w1 - 1, x2 + w2 - 1) - max(x1, x2) + 1
        ih = min(y1 + h1 - 1, y2 + h2 - 1) - max(y1, y2) + 1
        return iw * ih

    return 0

def get_monitor_area(search, monitors):
    """
    Returns the monitor with the most overlap with the 'search' rectangle.

    :param search: A 4-tuple representing a rectangle:
                   (top_left_x, top_left_y, width, height)
    :param monitors: A list of 4-tuples representing each monitor's rectangle.
    :return: The monitor rectangle with the most overlap with 'search'.
    :rtype: (top_left_x, top_left_y, width, height)
    """
    marea = 0
    mon = None
    for mx, my, mw, mh in monitors:
        a = rect_intersect_area((mx, my, mw, mh), search)
        if a > marea:
            marea = a
            mon = (mx, my, mw, mh)

    return mon

def monitor_rects(monitors):
    """
    Takes a list of monitors returned by ``xinerama.get_monitors`` and returns
    a new list of rectangles, in the same order, of monitor areas that account
    for all struts set by all windows. Duplicate struts are ignored.

    :param monitors: A list of 4-tuples representing monitor rectangles.
    :return: A list of 4-tuples representing monitor rectangles after
             subtracting strut areas.
    :rtype: [(top_left_x, top_left_y, width, height)]
    """
    mons = monitors # alias
    wa = mons[:]

    clients = ewmh.get_client_list().reply()

    log = [] # Identical struts should be ignored

    for c in clients:
        try:
            cx, cy, cw, ch = window.get_geometry(c)
        except xproto.BadWindow:
            continue

        for i, (x, y, w, h) in enumerate(wa):
            if rect_intersect_area((x, y, w, h), (cx, cy, cw, ch)) > 0:
                struts = ewmh.get_wm_strut_partial(c).reply()
                if not struts:
                    struts = ewmh.get_wm_strut(c).reply()

                key = (cx, cy, cw, ch, struts)
                if key in log:
                    continue
                log.append(key)

                if struts and not all([v == 0 for v in struts.itervalues()]):
                    if struts['left'] or struts['right']:
                        if struts['left']:
                            x += cw
                        w -= cw
                    if struts['top'] or struts['bottom']:
                        if struts['top']:
                            y += ch
                        h -= ch
                elif struts:
                    # x/y shouldn't be zero
                    if cx > 0 and w == cx + cw:
                        w -= cw
                    elif cy > 0 and h == cy + ch:
                        h -= ch
                    elif cx > 0 and x == cx:
                        x += cw
                        w -= cw
                    elif cy > 0 and y == cy:
                        y += ch
                        h -= ch

                wa[i] = (x, y, w, h)

    return wa




from __future__ import print_function
import sys

from setuptools import setup

try:
    from xpybutil.compat import xproto, xinerama, randr
except:
    print('')
    print('xpybutil requires the X Python Binding')
    print('See: http://cgit.freedesktop.org/xcb/xpyb/')
    print('More options: xpyb-ng:', 'https://github.com/dequis/xpyb-ng',
          'and xcffib:', 'https://github.com/tych0/xcffib')
    sys.exit(1)

setup(
    name="xpybutil",
    maintainer="Fenner Macrae",
    author_email="fmacrae.dev@gmail.com",
    version="0.0.5",
    license="WTFPL",
    description="An incomplete xcb-util port plus some extras",
    long_description="See README",
    url="http://pytyle.com",
    platforms='POSIX',
    packages=['xpybutil'],
    data_files=[('share/doc/xpybutil', ['README.md', 'COPYING'])]
)



"""
Implements a subset of the Motif spec. This module exists
because some window managers still use this as the only way of
toggling window decorations via events.
"""
import struct

from xpybutil.compat import xproto

from xpybutil import conn as c
from xpybutil import util

__atoms = ['_MOTIF_WM_HINTS']

class Hint:
    Functions = 1 << 0
    Decorations = 1 << 1
    InputMode = 1 << 2
    Status = 1 << 3

class Function:
    _None = 0
    All = 1 << 0
    Resize = 1 << 1
    Move = 1 << 2
    Minimize = 1 << 3
    Maximize = 1 << 4
    Close = 1 << 5

class Decoration:
    _None = 0
    All = 1 << 0
    Border = 1 << 1
    ResizeH = 1 << 2
    Title = 1 << 3
    Menu = 1 << 4
    Minimize = 1 << 5
    Maximize = 1 << 6

class Input:
    Modeless = 0
    PrimaryApplicationModal = 1
    SystemModal = 2
    FullApplicationModal = 3

class Status:
    TearoffWindow = 1 << 0

# Some aliases
atom = util.get_atom
preplace = xproto.PropMode.Replace

# Build the atom cache for quicker access
util.build_atom_cache(__atoms)

# _MOTIF_WM_HINTS

class MotifHintsCookie(util.PropertyCookie):
    def reply(self):
        v = util.PropertyCookie.reply(self)

        if not v:
            return None

        return {
            'flags': {
                'Functions': v[0] & Hint.Functions > 0,
                'Decorations': v[0] & Hint.Decorations > 0,
                'InputMode': v[0] & Hint.InputMode > 0,
                'Status': v[0] & Hint.Status > 0
            },
            'function': v[1],
            'decoration': v[2],
            'input': v[3],
            'status': v[4]
        }

def get_hints(window):
    return MotifHintsCookie(util.get_property(window, '_MOTIF_WM_HINTS'))

def get_hints_unchecked(window):
    cook = util.get_property_unchecked(window, '_MOTIF_WM_HINTS')
    return MotifHintsCookie(cook)

def _pack_hints(flags, function, decoration, input, status):
    hints = [0] * 5

    if flags & Hint.Functions:
        hints[1] = function

    if flags & Hint.Decorations:
        hints[2] = decoration

    if flags & Hint.InputMode:
        hints[3] = input

    if flags & Hint.Status:
        hints[4] = status

    hints[0] = flags

    return struct.pack('I' * 5, *hints)

def set_hints(window, flags, function=Function._None,
              decoration=Decoration._None, input=Input.Modeless,
              status=Status.TearoffWindow):
    packed = _pack_hints(flags, function, decoration, input, status)
    return c.core.ChangeProperty(preplace, window, atom('_MOTIF_WM_HINTS'),
                                 atom('_MOTIF_WM_HINTS'), 32, 5, packed)

def set_hints_checked(window, flags, function=Function._None,
                      decoration=Decoration._None, input=Input.Modeless,
                      status=Status.TearoffWindow):
    packed = _pack_hints(flags, function, decoration, input, status)
    return c.core.ChangePropertyChecked(preplace, window,
                                        atom('_MOTIF_WM_HINTS'),
                                        atom('_MOTIF_WM_HINTS'), 32, 5, packed)




#!/usr/bin/env python2

import argparse
import sys

from xpybutil.compat import xproto

import xpybutil
import xpybutil.event as event
import xpybutil.ewmh as ewmh
import xpybutil.keybind as keybind

epilog = '''
Using window-marker is exactly like using marks in vim. Namely, by pressing the
keybinding for "mark", window-marker then listens for one more letter (a-z) to
be entered. Whichever letter is entered will now be associated with the active
window.
If you now want to go back to that window (even if you're on a different
desktop), simply press the keybinding for "goto" and then enter the letter you
used to mark the window.
To set your own keybinding, please consult /usr/include/X11/keysymdef.h
and the output of `xmodmap`. The `xev` program may also be useful.
'''
parser = argparse.ArgumentParser(description='Vim-like marks for windows.',
                                 epilog=epilog)
parser.add_argument('-m', '--mark', default='Mod4-m', metavar='KEYBINDING',
                    help='keybinding to add a mark (default: %(default)s)')
parser.add_argument('-g', '--goto', default='Mod4-apostrophe',
                    metavar='KEYBINDING',
                    help='keybinding to goto a mark (default %(default)s)')
args = parser.parse_args()

keybinds = { args.mark: 'mark_window', args.goto: 'goto_window' }
keybindmap = {}
marked = {}
grabbing = None

def do_mark_window(letter):
    awin = ewmh.get_active_window().reply()
    if awin is not None:
        marked[letter] = awin

def do_goto_window(letter):
    if letter not in marked:
        print >> sys.stderr, 'mark %s does not exist' % letter
        return

    wid = marked[letter]
    try:
        wdesk = ewmh.get_wm_desktop(wid).reply()
        desktop = ewmh.get_current_desktop().reply()
        visibles = ewmh.get_visible_desktops().reply() or [desktop]

        if wdesk is not None and wdesk not in visibles:
            ewmh.request_current_desktop_checked(wdesk).check()
        ewmh.request_active_window_checked(wid, source=1).check()
    except xproto.BadWindow:
        print >> sys.stderr, '%d no longer exists' % wid

def mark_window():
    start_get_letter(do_mark_window)

def goto_window():
    start_get_letter(do_goto_window)

def start_get_letter(cb):
    global grabbing

    GS = xproto.GrabStatus
    if keybind.grab_keyboard(xpybutil.root).status == GS.Success:
        grabbing = cb

def cb_get_letter(e):
    global grabbing

    if grabbing is not None:
        keybind.ungrab_keyboard()
        sym = keybind.get_keysym(e.detail)
        letter = keybind.get_keysym_string(sym)

        if len(letter) == 1 and ord(letter) in range(ord('a'), ord('z') + 1):
            grabbing(letter.lower())

        grabbing = None

# This has to come first so it is called first in the event loop
event.connect('KeyPress', xpybutil.root, cb_get_letter)

for key_str, fun_str in keybinds.iteritems():
    if fun_str not in globals():
        print >> sys.stderr, 'No such function %s for %s' % (fun_str, key_str)
        continue

    fun = globals()[fun_str]
    if not keybind.bind_global_key('KeyPress', key_str, fun):
        print >> sys.stderr, 'Could not bind %s to %s' % (key_str, fun_str)

event.main()




"""
This module is pretty inadequate. Mostly because if you find
yourself wanting to use this module, you're probably doing something wrong.
(i.e., use pycairo or PIL.)

This module will likely be removed in the future.
"""
from xpybutil.compat import xproto

def get_font_height(qfont):
    return qfont.max_bounds.ascent + qfont.max_bounds.descent

def get_text_width(qfont, text):
    assert isinstance(qfont, xproto.QueryFontReply)

    return sum([qfont.char_infos[ord(i)].character_width for i in text])

def get_text_height(qfont, text):
    assert isinstance(qfont, xproto.QueryFontReply)

    cinfo = qfont.char_infos
    return max([cinfo[ord(i)].ascent + cinfo[ord(i)].descent for i in text])




"""
A set of functions devoted to binding key presses and registering
callbacks. This will automatically hook into the event callbacks
in event.py.

The two functions of interest here are 'bind_global_key' and 'bind_key'. Most
of the other functions facilitate the use of those two, but you may need them
if you're getting down and dirty.
"""
from collections import defaultdict
import sys

from xpybutil.compat import xproto

from xpybutil import conn, root, event
from xpybutil.keysymdef import keysyms, keysym_strings

__kbmap = None
__keysmods = None

__keybinds = defaultdict(list)
__keygrabs = defaultdict(int) # Key grab key -> number of grabs

EM = xproto.EventMask
GM = xproto.GrabMode
TRIVIAL_MODS = [
    0,
    xproto.ModMask.Lock,
    xproto.ModMask._2,
    xproto.ModMask.Lock | xproto.ModMask._2
]

def bind_global_key(event_type, key_string, cb):
    """
    An alias for ``bind_key(event_type, ROOT_WINDOW, key_string, cb)``.

    :param event_type: Either 'KeyPress' or 'KeyRelease'.
    :type event_type: str
    :param key_string: A string of the form 'Mod1-Control-a'.
                       Namely, a list of zero or more modifiers separated by
                       '-', followed by a single non-modifier key.
    :type key_string: str
    :param cb: A first class function with no parameters.
    :type cb: function
    :return: True if the binding was successful, False otherwise.
    :rtype: bool
    """
    return bind_key(event_type, root, key_string, cb)

def bind_key(event_type, wid, key_string, cb):
    """
    Binds a function ``cb`` to a particular key press ``key_string`` on a
    window ``wid``. Whether it's a key release or key press binding is
    determined by ``event_type``.

    ``bind_key`` will automatically hook into the ``event`` module's dispatcher,
    so that if you're using ``event.main()`` for your main loop, everything
    will be taken care of for you.

    :param event_type: Either 'KeyPress' or 'KeyRelease'.
    :type event_type: str
    :param wid: The window to bind the key grab to.
    :type wid: int
    :param key_string: A string of the form 'Mod1-Control-a'.
                       Namely, a list of zero or more modifiers separated by
                       '-', followed by a single non-modifier key.
    :type key_string: str
    :param cb: A first class function with no parameters.
    :type cb: function
    :return: True if the binding was successful, False otherwise.
    :rtype: bool
    """
    assert event_type in ('KeyPress', 'KeyRelease')

    mods, kc = parse_keystring(key_string)
    key = (wid, mods, kc)

    if not kc:
        print >> sys.stderr, 'Could not find a keycode for %s' % key_string
        return False

    if not __keygrabs[key] and not grab_key(wid, mods, kc):
        return False

    __keybinds[key].append(cb)
    __keygrabs[key] += 1

    if not event.is_connected(event_type, wid, __run_keybind_callbacks):
        event.connect(event_type, wid, __run_keybind_callbacks)

    return True

def parse_keystring(key_string):
    """
    A utility function to turn strings like 'Mod1-Mod4-a' into a pair
    corresponding to its modifiers and keycode.

    :param key_string: String starting with zero or more modifiers followed
                       by exactly one key press.

                       Available modifiers: Control, Mod1, Mod2, Mod3, Mod4,
                       Mod5, Shift, Lock
    :type key_string: str
    :return: Tuple of modifier mask and keycode
    :rtype: (mask, int)
    """
    modifiers = 0
    keycode = None

    for part in key_string.split('-'):
        if hasattr(xproto.KeyButMask, part):
            modifiers |= getattr(xproto.KeyButMask, part)
        else:
            if len(part) == 1:
                part = part.lower()
            keycode = lookup_string(part)

    return modifiers, keycode

def lookup_string(kstr):
    """
    Finds the keycode associated with a string representation of a keysym.

    :param kstr: English representation of a keysym.
    :return: Keycode, if one exists.
    :rtype: int
    """
    if kstr in keysyms:
        return get_keycode(keysyms[kstr])
    elif len(kstr) > 1 and kstr.capitalize() in keysyms:
        return get_keycode(keysyms[kstr.capitalize()])

    return None

def lookup_keysym(keysym):
    """
    Finds the english string associated with a keysym.

    :param keysym: An X keysym.
    :return: English string representation of a keysym.
    :rtype: str
    """
    return get_keysym_string(keysym)

def get_min_max_keycode():
    """
    Return a tuple of the minimum and maximum keycode allowed in the
    current X environment.

    :rtype: (int, int)
    """
    return conn.get_setup().min_keycode, conn.get_setup().max_keycode

def get_keyboard_mapping():
    """
    Return a keyboard mapping cookie that can be used to fetch the table of
    keysyms in the current X environment.

    :rtype: xcb.xproto.GetKeyboardMappingCookie
    """
    mn, mx = get_min_max_keycode()

    return conn.core.GetKeyboardMapping(mn, mx - mn + 1)

def get_keyboard_mapping_unchecked():
    """
    Return an unchecked keyboard mapping cookie that can be used to fetch the
    table of keysyms in the current X environment.

    :rtype: xcb.xproto.GetKeyboardMappingCookie
    """
    mn, mx = get_min_max_keycode()

    return conn.core.GetKeyboardMappingUnchecked(mn, mx - mn + 1)

def get_keysym(keycode, col=0, kbmap=None):
    """
    Get the keysym associated with a particular keycode in the current X
    environment. Although we get a list of keysyms from X in
    'get_keyboard_mapping', this list is really a table with
    'keysys_per_keycode' columns and ``mx - mn`` rows (where ``mx`` is the
    maximum keycode and ``mn`` is the minimum keycode).

    Thus, the index for a keysym given a keycode is:
    ``(keycode - mn) * keysyms_per_keycode + col``.

    In most cases, setting ``col`` to 0 will work.

    Witness the utter complexity:
    http://tronche.com/gui/x/xlib/input/keyboard-encoding.html

    You may also pass in your own keyboard mapping using the ``kbmap``
    parameter, but xpybutil maintains an up-to-date version of this so you
    shouldn't have to.

    :param keycode: A physical key represented by an integer.
    :type keycode: int
    :param col: The column in the keysym table to use.
                Unless you know what you're doing, just use 0.
    :type col: int
    :param kbmap: The keyboard mapping to use.
    :type kbmap: xcb.xproto.GetKeyboardMapingReply
    """
    if kbmap is None:
        kbmap = __kbmap

    mn, mx = get_min_max_keycode()
    per = kbmap.keysyms_per_keycode
    ind = (keycode - mn) * per + col

    return kbmap.keysyms[ind]

def get_keysym_string(keysym):
    """
    A simple wrapper to find the english string associated with a particular
    keysym.

    :param keysym: An X keysym.
    :rtype: str
    """
    return keysym_strings.get(keysym, [None])[0]

def get_keycode(keysym):
    """
    Given a keysym, find the keycode mapped to it in the current X environment.
    It is necessary to search the keysym table in order to do this, including
    all columns.

    :param keysym: An X keysym.
    :return: A keycode or None if one could not be found.
    :rtype: int
    """
    mn, mx = get_min_max_keycode()
    cols = __kbmap.keysyms_per_keycode
    for i in range(mn, mx + 1):
        for j in range(0, cols):
            ks = get_keysym(i, col=j)
            if ks == keysym:
                return i

    return None

def get_mod_for_key(keycode):
    """
    Finds the modifier that is mapped to the given keycode.
    This may be useful when analyzing key press events.

    :type keycode: int
    :return: A modifier identifier.
    :rtype: xcb.xproto.ModMask
    """
    return __keysmods.get(keycode, 0)

def get_keys_to_mods():
    """
    Fetches and creates the keycode -> modifier mask mapping. Typically, you
    shouldn't have to use this---xpybutil will keep this up to date if it
    changes.

    This function may be useful in that it should closely replicate the output
    of the ``xmodmap`` command. For example:

     ::

        keymods = get_keys_to_mods()
        for kc in sorted(keymods, key=lambda kc: keymods[kc]):
            print keymods[kc], hex(kc), get_keysym_string(get_keysym(kc))

    Which will very closely replicate ``xmodmap``. I'm not getting precise
    results quite yet, but I do believe I'm getting at least most of what
    matters. (i.e., ``xmodmap`` returns valid keysym strings for some that
    I cannot.)

    :return: A dict mapping from keycode to modifier mask.
    :rtype: dict
    """
    mm = xproto.ModMask
    modmasks = [mm.Shift, mm.Lock, mm.Control,
                mm._1, mm._2, mm._3, mm._4, mm._5] # order matters

    mods = conn.core.GetModifierMapping().reply()

    res = {}
    keyspermod = mods.keycodes_per_modifier
    for mmi in range(0, len(modmasks)):
        row = mmi * keyspermod
        for kc in mods.keycodes[row:row + keyspermod]:
            res[kc] = modmasks[mmi]

    return res

def get_modifiers(state):
    """
    Takes a ``state`` (typically found in key press or button press events)
    and returns a string list representation of the modifiers that were pressed
    when generating the event.

    :param state: Typically from ``some_event.state``.
    :return: List of modifier string representations.
    :rtype: [str]
    """
    ret = []

    if state & xproto.ModMask.Shift:
        ret.append('Shift')
    if state & xproto.ModMask.Lock:
        ret.append('Lock')
    if state & xproto.ModMask.Control:
        ret.append('Control')
    if state & xproto.ModMask._1:
        ret.append('Mod1')
    if state & xproto.ModMask._2:
        ret.append('Mod2')
    if state & xproto.ModMask._3:
        ret.append('Mod3')
    if state & xproto.ModMask._4:
        ret.append('Mod4')
    if state & xproto.ModMask._5:
        ret.append('Mod5')
    if state & xproto.KeyButMask.Button1:
        ret.append('Button1')
    if state & xproto.KeyButMask.Button2:
        ret.append('Button2')
    if state & xproto.KeyButMask.Button3:
        ret.append('Button3')
    if state & xproto.KeyButMask.Button4:
        ret.append('Button4')
    if state & xproto.KeyButMask.Button5:
        ret.append('Button5')

    return ret

def grab_keyboard(grab_win):
    """
    This will grab the keyboard. The effect is that further keyboard events
    will *only* be sent to the grabbing client. (i.e., ``grab_win``).

    N.B. There is an example usage of this in examples/window-marker.

    :param grab_win: A window identifier to report keyboard events to.
    :type grab_win: int
    :rtype: xcb.xproto.GrabStatus
    """
    return conn.core.GrabKeyboard(False, grab_win, xproto.Time.CurrentTime,
                                  GM.Async, GM.Async).reply()

def ungrab_keyboard():
    """
    This will release a grab initiated by ``grab_keyboard``.

    :rtype: void
    """
    conn.core.UngrabKeyboardChecked(xproto.Time.CurrentTime).check()

def grab_key(wid, modifiers, key):
    """
    Grabs a key for a particular window and a modifiers/key value.
    If the grab was successful, return True. Otherwise, return False.
    If your client is grabbing keys, it is useful to notify the user if a
    key wasn't grabbed. Keyboard shortcuts not responding is disorienting!

    Also, this function will grab several keys based on varying modifiers.
    Namely, this accounts for all of the "trivial" modifiers that may have
    an effect on X events, but probably shouldn't effect key grabbing. (i.e.,
    whether num lock or caps lock is on.)

    N.B. You should probably be using 'bind_key' or 'bind_global_key' instead.

    :param wid: A window identifier.
    :type wid: int
    :param modifiers: A modifier mask.
    :type modifiers: int
    :param key: A keycode.
    :type key: int
    :rtype: bool
    """
    try:
        for mod in TRIVIAL_MODS:
            conn.core.GrabKeyChecked(True, wid, modifiers | mod, key, GM.Async,
                                     GM.Async).check()

        return True
    except xproto.BadAccess:
        return False

def ungrab_key(wid, modifiers, key):
    """
    Ungrabs a key that was grabbed by ``grab_key``. Similarly, it will return
    True on success and False on failure.

    When ungrabbing a key, the parameters to this function should be
    *precisely* the same as the parameters to ``grab_key``.

    :param wid: A window identifier.
    :type wid: int
    :param modifiers: A modifier mask.
    :type modifiers: int
    :param key: A keycode.
    :type key: int
    :rtype: bool
    """
    try:
        for mod in TRIVIAL_MODS:
            conn.core.UngrabKeyChecked(key, wid, modifiers | mod).check()

        return True
    except xproto.BadAccess:
        return False

def update_keyboard_mapping(e):
    """
    Whenever the keyboard mapping is changed, this function needs to be called
    to update xpybutil's internal representing of the current keysym table.
    Indeed, xpybutil will do this for you automatically.

    Moreover, if something is changed that affects the current keygrabs,
    xpybutil will initiate a regrab with the changed keycode.

    :param e: The MappingNotify event.
    :type e: xcb.xproto.MappingNotifyEvent
    :rtype: void
    """
    global __kbmap, __keysmods

    newmap = get_keyboard_mapping().reply()

    if e is None:
        __kbmap = newmap
        __keysmods = get_keys_to_mods()
        return

    if e.request == xproto.Mapping.Keyboard:
        changes = {}
        for kc in range(*get_min_max_keycode()):
            knew = get_keysym(kc, kbmap=newmap)
            oldkc = get_keycode(knew)
            if oldkc != kc:
                changes[oldkc] = kc

        __kbmap = newmap
        __regrab(changes)
    elif e.request == xproto.Mapping.Modifier:
        __keysmods = get_keys_to_mods()

def __run_keybind_callbacks(e):
    """
    A private function that intercepts all key press/release events, and runs
    their corresponding callback functions. Nothing much to see here, except
    that we must mask out the trivial modifiers from the state in order to
    find the right callback.

    Callbacks are called in the order that they have been added. (FIFO.)

    :param e: A Key{Press,Release} event.
    :type e: xcb.xproto.Key{Press,Release}Event
    :rtype: void
    """
    kc, mods = e.detail, e.state
    for mod in TRIVIAL_MODS:
        mods &= ~mod

    key = (e.event, mods, kc)
    for cb in __keybinds.get(key, []):
        try:
            cb(e)
        except TypeError:
            cb()

def __regrab(changes):
    """
    Takes a dictionary of changes (mapping old keycode to new keycode) and
    regrabs any keys that have been changed with the updated keycode.

    :param changes: Mapping of changes from old keycode to new keycode.
    :type changes: dict
    :rtype: void
    """
    for wid, mods, kc in __keybinds.keys():
        if kc in changes:
            ungrab_key(wid, mods, kc)
            grab_key(wid, mods, changes[kc])

            old = (wid, mods, kc)
            new = (wid, mods, changes[kc])
            __keybinds[new] = __keybinds[old]
            del __keybinds[old]

if conn is not None:
    update_keyboard_mapping(None)
    event.connect('MappingNotify', None, update_keyboard_mapping)

